/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $
 * $Date: 2011/02/16 17:45:21 $
 * $Revision: 1.102 $
 * $Name: HEAD $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 * @defgroup create_master    Create Master Files (xsh_create_master_dark/bias/flat)
 * @ingroup drl_functions
 *
 * 
 *
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
  Includes
  -----------------------------------------------------------------------------*/
#include <xsh_drl.h>
#include <xsh_dfs.h>
//#include <xsh_pfits.h>
#include <xsh_pfits_qc.h>
#include <xsh_utils_wrappers.h>
#include <xsh_pfits.h>
#include <xsh_pfits_qc.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <xsh_badpixelmap.h>
#include <xsh_paf_save.h>
#include <xsh_utils_image.h>
#include <xsh_detmon.h>  
#include <xsh_irplib_utils.h>  //For skip_if
#include <math.h>

/*-----------------------------------------------------------------------------
  Typedefs
  ---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
  Defines
  ---------------------------------------------------------------------------*/

#define RECIPE_ID "xsh_mbias"
#define HIST_FACT 2.354820045
/*-----------------------------------------------------------------------------
  Functions prototypes
  -----------------------------------------------------------------------------*/

static cpl_error_code
xsh_compute_ron_mbias(cpl_frameset* raws,
		      xsh_pre* master,
		      cpl_parameterlist* parameters);


static cpl_error_code
xsh_compute_fpn_mbias(cpl_frameset* raws,
		      xsh_pre* master,
		      xsh_instrument* instrument,
		      cpl_parameterlist* parameters);

static cpl_error_code
xsh_mbias_get_fpn(const cpl_image* ima,
                  cpl_parameterlist* parameters,
                  double* fpn);


static cpl_error_code
xsh_my_detmon_ronbias_random(const cpl_imagelist * rawbiases,
			     const cpl_image     * masterbias,
                             cpl_propertylist    * qclist,
                             const char* method,
                             const int random_sizex,
                             const int random_nsamples);

static cpl_error_code
xsh_my_detmon_ronbias_histo(const cpl_imagelist * rawbiases,
			 const cpl_image * masterbias,
			 cpl_propertylist * qclist,
			          const char* method);

static cpl_error_code
xsh_my_detmon_ronbias_region(const cpl_imagelist * rawbiases,
			     const cpl_image * masterbias,
                             cpl_propertylist * qclist,
                             const char* pmethod,
                             const int ref_llx,
                             const int ref_lly,
                             const int ref_urx,
                             const int ref_ury,
                             const int stacking_ks_low,
			     const int stacking_ks_iter);



static cpl_error_code
xsh_my_detmon_ronbias_preoverscan(const cpl_imagelist * rawbiases,
//				  const cpl_image * masterbias,
                                  cpl_propertylist * qclist,
                                  cpl_image ** synthetic,
                                  const int prescan_llx,
                                  const int prescan_lly,
                                  const int prescan_urx,
                                  const int prescan_ury,
                                  const int overscan_llx,
                                  const int overscan_lly,
                                  const int overscan_urx,
                                  const int overscan_ury,
                                  const int stacking_ks_low,
                                  const int stacking_ks_iter);




/*-----------------------------------------------------------------------------
  Implementation
  -----------------------------------------------------------------------------*/


static double
xsh_mdark_get_contam(xsh_pre * pre,cpl_parameterlist* drs_params,
xsh_instrument* instr)
{

   double contam=0;

   int llx_on=0;
   int lly_on=0;
   int urx_on=0;
   int ury_on=0;

   int llx_off=0;
   int lly_off=0;
   int urx_off=0;
   int ury_off=0;
   if(xsh_parameters_find(drs_params,"xsh_mdark","contam_on_llx")!=NULL){
      check(llx_on=xsh_parameters_get_int(drs_params,"xsh_mdark","contam_on_llx"));
      check(lly_on=xsh_parameters_get_int(drs_params,"xsh_mdark","contam_on_lly"));
      check(urx_on=xsh_parameters_get_int(drs_params,"xsh_mdark","contam_on_urx"));
      check(ury_on=xsh_parameters_get_int(drs_params,"xsh_mdark","contam_on_ury"));


      check(llx_off=xsh_parameters_get_int(drs_params,"xsh_mdark","contam_off_llx"));
      check(lly_off=xsh_parameters_get_int(drs_params,"xsh_mdark","contam_off_lly"));
      check(urx_off=xsh_parameters_get_int(drs_params,"xsh_mdark","contam_off_urx"));
      check(ury_off=xsh_parameters_get_int(drs_params,"xsh_mdark","contam_off_ury"));

   } else {
      if(xsh_instrument_get_arm(instr) == XSH_ARM_UVB) {
         llx_on =801; lly_on =1058;urx_on =810; ury_on =1067;
         llx_off=855; lly_off=1066;urx_off=864; ury_off=1075;
      } else if(xsh_instrument_get_arm(instr) == XSH_ARM_VIS) {
         llx_on =1672; lly_on =2672;urx_on =1681;ury_on =2681;
         llx_off=1926; lly_off=2941;urx_off=1935;ury_off=2950;
      } else if(xsh_instrument_get_arm(instr) == XSH_ARM_NIR) {
         llx_on =991; lly_on =1083;urx_on =1015;ury_on =1108;
         llx_off=1029;lly_off=1075;urx_off=1053;ury_off=1110;
      }
   }

   llx_on=llx_on/pre->binx;
   lly_on=lly_on/pre->biny;
   urx_on=urx_on/pre->binx;
   ury_on=ury_on/pre->biny;

   llx_off=llx_off/pre->binx;
   lly_off=lly_off/pre->biny;
   urx_off=urx_off/pre->binx;
   ury_off=ury_off/pre->biny;

   check(contam=
         cpl_image_get_median_window(pre->data,llx_on,lly_on,urx_on,ury_on)-
         cpl_image_get_median_window(pre->data,llx_off,lly_off,urx_off,ury_off));


cleanup:

 
  return contam;


} 


/** 
 * Calculates and sets QC parameters: Median, Average, Stdev and Slope, for
 * primary image AND qual.
 * 
 * @param preFrame Master Dark Frame (PRE format)
 * @param instrument Pointer to the instrument description structure
 * @param parameters parameters of recipe
 * @param drs_params parameters specified by DRS setup table
 * @param crh_frm CRH frame
 * @param bpm_frm BPM frame 
 */
static void
set_masterdark_qc (xsh_pre * preFrame, 
                   xsh_instrument * instrument,
                   cpl_parameterlist* drs_params,
                   cpl_parameterlist* parameters,
                   double exptime,
                   cpl_frame* crh_frm, cpl_frame* bpm_frm)
{
  int nx =0, ny = 0;
  double avg_value = 0.0;
  double median_value = 0.0;
  double stdev = 0.0;
  cpl_parameter* p=NULL;

  int ref_llx=-1;
  int ref_lly=-1;
  int ref_urx=-1;
  int ref_ury=-1;


  int ron_llx=-1;
  int ron_lly=-1;
  int ron_urx=-1;
  int ron_ury=-1;
  int ron_nsamp=100;
  int ron_hsize=4;

  int zone_fpn[4];
  int fpn_llx=-1;
  int fpn_lly=-1;
  int fpn_urx=-1;
  int fpn_ury=-1;
  int fpn_nsamp=100;
  int fpn_hsize=4;


  double qc_fpn_val=0;
  double qc_fpn_err=0;
  double contam=0;

  int sx=0;
  int sy=0;
  const char* crh_name=NULL;
  const char* bpm_name=NULL;
  cpl_image* crh_ima=NULL;
  cpl_image* bpm_ima=NULL;
  int* pcrh=NULL;
  int* pbpm=NULL;
  int* pqua=NULL;
  int i=0;
  xsh_pre* dup_frm=NULL;
  bool bpm_found=false;
  bool crh_found=false;


  nx = xsh_pre_get_nx (preFrame);
  assure (nx != 0, cpl_error_get_code (), "Cant get X size");
  ny = xsh_pre_get_ny (preFrame);
  assure (ny != 0, cpl_error_get_code (), "Cant get Y size");

  sx=cpl_image_get_size_x(preFrame->data);
  sy=cpl_image_get_size_y(preFrame->data);
   if(drs_params !=NULL){
      check(ref_llx=xsh_parameters_get_int(drs_params,"xsh_mdark","ref1_llx"));
      check(ref_lly=xsh_parameters_get_int(drs_params,"xsh_mdark","ref1_lly"));
      check(ref_urx=xsh_parameters_get_int(drs_params,"xsh_mdark","ref1_urx"));
      check(ref_ury=xsh_parameters_get_int(drs_params,"xsh_mdark","ref1_ury"));
   } else {
      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ref1_llx"));
      check(ref_llx = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ref1_lly"));
      check(ref_lly = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ref1_urx"));
      check(ref_urx = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ref1_ury"));
      check(ref_ury = cpl_parameter_get_int(p));
   }
  /* QC parameters */

  if(ref_llx == -1) ref_llx=1;
  if(ref_lly == -1) ref_lly=1;  
  if(ref_urx == -1) ref_urx=sx;
  if(ref_ury == -1) ref_ury=sy;

  ref_llx = (ref_llx>0) ? ref_llx : 1;
  ref_lly = (ref_lly>0) ? ref_lly : 1;
  ref_urx = (ref_urx<sx) ? ref_urx : sx;
  ref_ury = (ref_ury<sy) ? ref_ury : sy;


   if(drs_params!=NULL){

      check(ron_llx=xsh_parameters_get_int(drs_params,"xsh_mdark","ron_llx"));
      check(ron_lly=xsh_parameters_get_int(drs_params,"xsh_mdark","ron_lly"));
      check(ron_urx=xsh_parameters_get_int(drs_params,"xsh_mdark","ron_urx"));
      check(ron_ury=xsh_parameters_get_int(drs_params,"xsh_mdark","ron_ury"));
      check(ron_hsize=xsh_parameters_get_int(drs_params,"xsh_mdark","ron_hsize"));
      check(ron_nsamp=xsh_parameters_get_int(drs_params,"xsh_mdark","ron_nsamples"));

   } else {

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ron_llx"));
      check(ron_llx = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ron_lly"));
      check(ron_lly = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ron_urx"));
      check(ron_urx = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ron_urx"));
      check(ron_ury = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ron_hsize"));
      check(ron_hsize = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ron_nsamples"));
      check(ron_nsamp = cpl_parameter_get_int(p));

   }
  /* QC parameters */
  if(ron_llx == -1) ron_llx=1;
  if(ron_lly == -1) ron_lly=1;  
  if(ron_urx == -1) ron_urx=sx;
  if(ron_ury == -1) ron_ury=sy;

  ron_llx = (ron_llx>0) ? ron_llx : 1;
  ron_lly = (ron_lly>0) ? ron_lly : 1;
  ron_urx = (ron_urx<sx) ? ron_urx : sx;
  ron_ury = (ron_ury<sy) ? ron_ury : sy;




   if(drs_params !=NULL){
      check(fpn_llx=xsh_parameters_get_int(drs_params,"xsh_mdark","fpn_llx"));
      check(fpn_lly=xsh_parameters_get_int(drs_params,"xsh_mdark","fpn_lly"));
      check(fpn_urx=xsh_parameters_get_int(drs_params,"xsh_mdark","fpn_urx"));
      check(fpn_ury=xsh_parameters_get_int(drs_params,"xsh_mdark","fpn_ury"));
   } else {
      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.fpn_llx"));
      check(fpn_llx = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.fpn_lly"));
      check(fpn_lly = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.fpn_urx"));
      check(fpn_urx = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.fpn_urx"));
      check(fpn_ury = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.fpn_hsize"));
      check(fpn_hsize = cpl_parameter_get_int(p));

      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.fpn_nsamples"));
      check(fpn_nsamp = cpl_parameter_get_int(p));
   }
  /* QC parameters */


  if(fpn_llx == -1) fpn_llx=1;
  if(fpn_lly == -1) fpn_lly=1;  
  if(fpn_urx == -1) fpn_urx=sx;
  if(fpn_ury == -1) fpn_ury=sy;

  fpn_llx = (fpn_llx>0) ? fpn_llx : 1;
  fpn_lly = (fpn_lly>0) ? fpn_lly : 1;
  fpn_urx = (fpn_urx<sx) ? fpn_urx : sx;
  fpn_ury = (fpn_ury<sy) ? fpn_ury : sy;




  /*
    Median and avg should ignore the bad pixels including the Cosmics.
    But at this point the preframe->qual DOES NOT include the COSMIC RAYS !
    (see xsh_remove_crh_multi function in xsh_remove_crh_multi.c)
  */

  //check( xsh_pre_median_mean_stdev( preFrame, &mean, &median,
  //                                  &stdev ) ) ;
  dup_frm=xsh_pre_duplicate(preFrame);

  if(crh_frm!=NULL) {
    check(crh_name=cpl_frame_get_filename(crh_frm));
    check(crh_ima=cpl_image_load(crh_name,XSH_PRE_QUAL_TYPE,0,0));
    check(pcrh=cpl_image_get_data_int(crh_ima));
    crh_found=true;
  }

  if(bpm_frm != NULL) {
    check(bpm_name=cpl_frame_get_filename(bpm_frm));
    check(bpm_ima=cpl_image_load(bpm_name,XSH_PRE_QUAL_TYPE,0,0));
    check(pbpm=cpl_image_get_data_int(bpm_ima));
    bpm_found=true;
  }

  if(bpm_found || crh_found) {
    check(pqua=cpl_image_get_data_int( preFrame->qual));
    sx=preFrame->nx;
    sy=preFrame->ny;

    if(bpm_found && crh_found) {
      for(i=0;i<sx*sy;i++) {
	if(pcrh[i] != 0 || pbpm[i] != 0) pqua[i]=XSH_GOOD_PIXEL_LEVEL+1;
      }
    } else if (bpm_found) {
      for(i=0;i<sx*sy;i++) {
	if(pbpm[i] != 0) pqua[i]=XSH_GOOD_PIXEL_LEVEL+1;
      }
    } else if (crh_found) {
      for(i=0;i<sx*sy;i++) {
	if(pcrh[i] != 0) pqua[i]=XSH_GOOD_PIXEL_LEVEL+1;
      }
    }
  }
  xsh_free_image(&bpm_ima);
  xsh_free_image(&crh_ima);
  xsh_pre_free(&dup_frm);

  check( xsh_pre_median_mean_stdev_window( preFrame, 
                                           &avg_value, &median_value,&stdev,
                                           ref_llx,ref_lly,ref_urx,ref_ury) ) ;

  xsh_pfits_set_qc_mdarkavg( preFrame->data_header,avg_value) ;
  xsh_pfits_set_qc_mdarkmed( preFrame->data_header,median_value) ;
  xsh_pfits_set_qc_mdarkrms( preFrame->data_header,stdev) ;

  /*
    Must be done also for qual header, because the qual header is used
    for the paf file of the bpmap product
  */
  xsh_pfits_set_qc_mdarkavg( preFrame->qual_header,avg_value) ;
  xsh_pfits_set_qc_mdarkmed( preFrame->qual_header,median_value) ;
  xsh_pfits_set_qc_mdarkrms( preFrame->qual_header,stdev) ;

  /* FPN computation */
 

  zone_fpn[0]=fpn_llx;
  zone_fpn[1]=fpn_urx;
  zone_fpn[2]=fpn_lly;
  zone_fpn[3]=fpn_ury;
  check_msg(cpl_flux_get_noise_window(preFrame->data, zone_fpn, fpn_hsize,
				      fpn_nsamp, &qc_fpn_val, &qc_fpn_err),
	    "Error computing noise in a window");

  xsh_pfits_set_qc_norm_fpn(preFrame->qual_header,qc_fpn_val) ;
  xsh_pfits_set_qc_norm_fpn_err(preFrame->qual_header,qc_fpn_err) ;

  qc_fpn_val *= exptime;
  qc_fpn_err *= exptime;

  check(xsh_pfits_set_qc_fpn( preFrame->qual_header,qc_fpn_val) );

  check(contam=xsh_mdark_get_contam(preFrame,parameters,instrument));
  check(xsh_pfits_set_qc_contamination( preFrame->data_header,contam)) ;




  /*
    Should calculate the "slope", but how ?

    double slope = 0.0 ;
    slope = xsh_calculate_slope( preFrame->data ) ;

    xsh_pfits_set_qc( preFrame->data_header, (void *)&slope,
                      XSH_QC_MDARKSLOPE, instrument ) ;
  */

 cleanup:

  xsh_free_image(&bpm_ima);
  xsh_free_image(&crh_ima);
  xsh_pre_free(&dup_frm);


  return;
}

/** 
 * PELIMINARY:
 * Creates a master dark frame from input frame. Input frame is in PRE
 * format: medianFrame, errsFrame, bpmapFrame.
 * Divide median frame by exposure time and calculate QC params.
 * 
 * @param medFrame Frame in PRE format
 * @param instr instrument containing the arm in use
 * @param parameters of the recipe
 * @param drs_params parameters specified by DRS setup table
 * @param crh_frm CRH frame
 * @param bpm_frm BPM frame 
*/
cpl_frame *
xsh_create_master_dark (cpl_frame * medFrame, xsh_instrument* instr,
			cpl_parameterlist* parameters,
                        cpl_parameterlist* drs_params,
                        cpl_frame* crh_frm,cpl_frame* bpm_frm)
{
  /*
    Load frame into PRE 
    Modify KW, etc
    Save PRE and returns pointer to frame
  */
  xsh_pre *medPre = NULL;
  double exptime = 0.0;
  cpl_frame* resFrame = NULL;
  const  char* tag = XSH_GET_TAG_FROM_ARM (XSH_MASTER_DARK, instr);
  char* finalName = NULL;

  XSH_ASSURE_NOT_NULL(medFrame);
  XSH_ASSURE_NOT_NULL(instr);

  check(medPre = xsh_pre_load (medFrame,instr));
  XSH_ASSURE_NOT_NULL(medPre);

  exptime = medPre->exptime;

  /* If not NIR, Divide medFrameand medErrs by exposure time */
  if ( instr->arm != XSH_ARM_NIR ) {
    check_msg (cpl_image_divide_scalar (medPre->data, exptime),
	       "Cant divide median image by exptime %lf", exptime);
    check_msg (cpl_image_divide_scalar (medPre->errs, exptime),
	       "Cant divide median image by exptime %lf", exptime);

    /* Set the EXPTIME at 1. */
    xsh_msg_dbg_high ("Set EXPTIME = 1.");
    check(xsh_pfits_set_exptime (medPre->data_header, (double) 1.));
    check(xsh_pfits_set_exptime (medPre->errs_header, (double) 1.));
  }


 if ( xsh_instrument_get_arm(instr) == XSH_ARM_NIR ) {
   check(finalName =cpl_sprintf("masterdark_%s.fits", 
				xsh_instrument_arm_tostring(instr)));
 } else {
   /* Sabines doesn't like to have the binning in the file-name
   check(binx=xsh_pfits_get_binx(medPre->data_header));
   check(biny=xsh_pfits_get_biny(medPre->data_header));
   check(finalName =cpl_sprintf("masterdark_%s_%dx%d.fits", 
				xsh_instrument_arm_tostring(instr),
				binx,biny));
   */
   check(finalName =cpl_sprintf("masterdark_%s.fits", 
				xsh_instrument_arm_tostring(instr)));
 }
  /* Calculate and Save QC Keywords */
  /* QC.MDARKMED
     QC.MDARKRMS
     QC.MDARKSLOGE (?)
  */
  check(set_masterdark_qc (medPre, instr,drs_params,parameters,exptime,crh_frm,bpm_frm));


  /* Save PRE */

 

  check ( resFrame = xsh_pre_save (medPre, finalName, tag,1 ));
  assure (resFrame != NULL, cpl_error_get_code (), "Cant save PRE");
  xsh_msg_dbg_high ("PRE frame saved as file %s", finalName);

  check(cpl_frame_set_level (resFrame, CPL_FRAME_LEVEL_FINAL));
  check(cpl_frame_set_tag (resFrame, tag)); 

 cleanup:
  if (cpl_error_get_code() != CPL_ERROR_NONE){
    xsh_free_frame(&resFrame);
  }
  cpl_free(finalName);
  xsh_pre_free(&medPre);
  return resFrame;
}


/*---------------------------------------------------------------------------*/
/**
  @brief    Reject outlier pixels
  @param    image       image with pixels
  @param    min         reject pixels below this value
  @param    max         reject pixels abouve this value
 */
/*---------------------------------------------------------------------------*/
static void
reject_lo_hi(cpl_image *image, double min, double max)
{
  cpl_mask *mask_lo = NULL;
  cpl_mask *mask_hi = NULL;

  mask_lo = cpl_mask_threshold_image_create(image, -DBL_MAX, min);
  mask_hi = cpl_mask_threshold_image_create(image, max, DBL_MAX);
  assure_mem( mask_lo );
  assure_mem( mask_hi );

  cpl_mask_or(mask_lo, mask_hi);

  cpl_image_reject_from_mask(image, mask_lo);
  
  cleanup:
  xsh_free_mask(&mask_lo);
  xsh_free_mask(&mask_hi);
  return;
}

static int
count_good(const cpl_image *image)
{
    return 
        cpl_image_get_size_x(image) * cpl_image_get_size_y(image) - 
        cpl_image_count_rejected(image);
}

static double
get_masterbias_qc_structure_row_region(cpl_image * tima)
{

   cpl_image* avg_row=NULL;
   double min=0;
   double max=0;
   double struct_row=0;



  check(avg_row = cpl_image_collapse_create(tima,0));
  check(cpl_image_divide_scalar(avg_row,cpl_image_get_size_y(tima)));

  /* restricts statistics to +/- 2 ADU around mean */
  min = cpl_image_get_mean(avg_row) - 2;
  max = cpl_image_get_mean(avg_row) + 2;
 
  /* replace with MIDAS
  stat/ima avg_row + exc={min},{max};
  */
  check( reject_lo_hi(avg_row, min, max) );
  if (count_good(avg_row) >= 2)
      {
          check(struct_row = cpl_image_get_stdev(avg_row));
      }
  else
      {
          struct_row = -1;
          xsh_msg_warning("Only %d good pixels in image. Setting QC parameter to -1",
                           count_good(avg_row));
      }

  cleanup:
   xsh_free_image(&avg_row);
 

   return struct_row;

}




static double
get_masterbias_qc_structure_col_region(cpl_image * tima)
{


   cpl_image* avg_col=NULL;
   double min=0;
   double max=0;
   double struct_col=0;

   check(avg_col = cpl_image_collapse_create(tima,1));
   check(cpl_image_divide_scalar(avg_col,cpl_image_get_size_x(tima)));

   /* restricts statistics to +/- 2 ADU around mean */
   min = cpl_image_get_mean(avg_col) - 2;
   max = cpl_image_get_mean(avg_col) + 2; 

   /* replace with MIDAS
      stat/ima avg_col + exc={min},{max};
   */
   check( reject_lo_hi(avg_col, min, max) );
   if (count_good(avg_col) >= 2)
   {
      check(struct_col = cpl_image_get_stdev(avg_col));
   }
   else
   {
      struct_col = -1;
      xsh_msg_warning("Only %d good pixels in image. Setting QC parameter to -1",
                      count_good(avg_col));
   }

  cleanup:
   xsh_free_image(&avg_col);
 

   return struct_col;

}


/** 
 * Calculates and sets QC parameters: Median, Average, Stdev and Slope, for
 * primary image AND qual.
 * 
 * @param master Master Bias Frame (PRE format)
 * @param llx lower left X
 * @param lly lower left Y
 * @param urx upper right X
 * @param ury upper right Y
 * @param ref_x reference X
 * @param ref_y reference Y
 * @param reg_id region id
 * @param dlevel debug level
 * @return void
 */
/* not used
static void
set_masterbias_qc_xsh_structure_region(xsh_pre * master,
                                   const int llx,
                                   const int lly,
                                   const int urx,
                                   const int ury,
                                   const int ref_x,
                                   const int ref_y,
                                   const int reg_id,
                                   const int dlevel)
{

   double structx = 0.0;
   double structy = 0.0;

   if(reg_id==1){
      check(structx=get_masterbias_qc_xstructure_region(master,
                                                        llx,lly,ref_x,ury,
                                                        reg_id,dlevel));
   } else {
      check(structx=get_masterbias_qc_xstructure_region(master,
                                                        ref_x,lly,urx,ury,
                                                        reg_id,dlevel));
   }

   if(reg_id==1){
      check(structy=get_masterbias_qc_ystructure_region(master,
                                                        llx,lly,urx,ref_y,
                                                        reg_id,dlevel));
   } else {
      check(structy=get_masterbias_qc_ystructure_region(master,
                                                        llx,ref_y,urx,ury,
                                                        reg_id,dlevel));
   }


 cleanup:
   return;

}
*/

/** 
 * Calculates and sets QC parameters: Median, Average, Stdev and Slope, for
 * primary image AND qual.
 * 
 * @param master Master Bias Frame (PRE format)
 * @param llx lower left X
 * @param lly lower left Y
 * @param urx upper right X
 * @param ury upper right Y
 * @param ref_x reference X
 * @param ref_y reference Y
 * @param reg_id region id
 * @param dlevel debug level
 * @return void
 */
static void
set_masterbias_qc_uves_structure_region(xsh_pre * master,
                                   const int llx,
                                   const int lly,
                                   const int urx,
                                   const int ury,
                                   const int ref_x,
                                   const int ref_y,
                                   const int reg_id,
                                   const int dlevel)
{
   /* UVES algorithm */
   cpl_image* xima=NULL;
   cpl_image* yima=NULL;
   double struct_row=0;
   double struct_col=0;

   double mean = 0.0;
   double median = 0.0;
   double stdev = 0.0;

   double structx = 0.0;
   double structy = 0.0;
   int nb_images = 0;

   check( xsh_pre_median_mean_stdev_window( master, &mean, &median,&stdev,
                                            llx,lly,urx,ury) ) ;

   check(xsh_pfits_set_qc_mbiasavg (master->data_header, mean));
   check(xsh_pfits_set_qc_mbiasmed (master->data_header, median));
   check(xsh_pfits_set_qc_mbiasrms (master->data_header, stdev));



   if(reg_id==1){
      check(xima=cpl_image_extract(master->data,llx,lly,urx,ref_y));
   } else {
      check(xima=cpl_image_extract(master->data,llx,ref_y,urx,ury));
   }
   if (0) {
      /*
        replace/ima {mbia} {tmpfrm} 300,>=300.;
      */
      check(cpl_image_threshold(xima,
                                -DBL_MAX,median+3*stdev,
                                -DBL_MAX,median+3*stdev));
   }

   if(reg_id==1){
      check(yima=cpl_image_extract(master->data,llx,lly,ref_x,ury));
   } else {
      check(yima=cpl_image_extract(master->data,ref_x,lly,urx,ury));
   }
   if (0) {
      /*
        replace/ima {mbia} {tmpfrm} 300,>=300.;
      */
      check(cpl_image_threshold(yima,
                                -DBL_MAX,median+3*stdev,
                                -DBL_MAX,median+3*stdev));
   }

   check(struct_row=get_masterbias_qc_structure_row_region(xima));
   structx=struct_row;

   check(struct_col=get_masterbias_qc_structure_col_region(yima));
   structy=struct_col;

   /* end UVES based algorithm */
   if(reg_id==1) {
      check(xsh_pfits_set_qc_reg1_structx(master->data_header,structx));
      check(xsh_pfits_set_qc_reg1_structy(master->data_header,structy));
   } else {
      check(xsh_pfits_set_qc_reg2_structx(master->data_header,structx));
      check(xsh_pfits_set_qc_reg2_structy(master->data_header,structy));
   }

   check( nb_images = xsh_pfits_get_datancom( master->data_header));
   //AMo: this formula is inaccurate. RON is computed anyway elsewhere
   //check( xsh_pfits_set_qc_ron( master->data_header, 
   //  stdev*master->conad*sqrt(nb_images)));


  cleanup:
   xsh_free_image(&xima);
   xsh_free_image(&yima);
   return;

}

/** 
 * Calculates and sets QC parameters: Median, Average, Stdev and Slope, for
 * primary image AND qual.
 * 
 * @param master Master Bias Frame (PRE format)
 * @param llx lower left X
 * @param lly lower left Y
 * @param urx upper right X
 * @param ury upper right Y
 * @param ref_x reference X
 * @param ref_y reference Y
 * @param reg_id region id
 * @param dlevel debug level
 * @return void
 */
static void
set_masterbias_qc_structure_region(xsh_pre * master,
                                   const int llx,
                                   const int lly,
                                   const int urx,
                                   const int ury,
                                   const int ref_x,
                                   const int ref_y,
                                   const int reg_id,
                                   const int dlevel)
{


  /* Not really used: this computes a fit of bias structure's the slope
  check(set_masterbias_qc_xsh_structure_region(master,llx,lly,urx,ury,
					       ref_x,ref_y,reg_id,dlevel));
  */

  check(set_masterbias_qc_uves_structure_region(master,llx,lly,urx,ury,
					       ref_x,ref_y,reg_id,dlevel));

  cleanup:
   return;

}


/** 
 * Calculates and sets QC parameters: Median, Average, Stdev and Slope, for
 * primary image AND qual.
 * 
 * @param master Master Bias Frame (PRE format)
 * @param parameters Recipe input parameters
 */
static void
set_masterbias_qc_structure(xsh_pre * master, cpl_parameterlist* parameters)
{

  cpl_parameter* p=NULL;
  int ref_x=-1;
  int ref_y=-1;

  int ref_llx=0;
  int ref_lly=0;
  int ref_urx=0;
  int ref_ury=0;

  int sx=0;
  int sy=0;
  int dlevel=0;

  sx=cpl_image_get_size_x(master->data);
  sy=cpl_image_get_size_y(master->data);

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.struct_refx"));
  check(ref_x = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.struct_refy"));
  check(ref_y = cpl_parameter_get_int(p));

  check(dlevel=xsh_parameters_debug_level_get("xsh_mbias",parameters));

  ref_llx=1;ref_urx=sx;
  ref_lly=1;ref_ury=sy;

  if(ref_x == -1) {
     ref_x=sx/2;
  }

  if(ref_y == -1) {
     ref_y=sy/2;
  }

  ref_x = (ref_x>0) ? ref_x : 1;
  ref_y = (ref_y>0) ? ref_y : 1;
  ref_x = (ref_x<sx) ? ref_x : sx;
  ref_y = (ref_y<sy) ? ref_y : sy;

  set_masterbias_qc_structure_region(master,ref_llx,ref_lly,ref_urx,ref_ury,
				     ref_x,ref_y,1,dlevel);

 
  set_masterbias_qc_structure_region(master,ref_llx,ref_lly,ref_urx,ref_ury,
                                     ref_x,ref_y,2,dlevel);
  

  cleanup:
     return;

}




/*----------------------------------------------------------------------------*/


/** 
  @brief 
    Creates a master bias frame
  @param raws 
    The raw frames
  @param frame 
    The median frame
  @param instr 
    The instrument containing the arm in use
  @param parameters
    The instrument recipe parameters

  @return
    The master bias frame with QC parameters and PROCATG
    Comment: why compute the QC parameters inside this function?
             would not be cleaner to use a different function for 
             that purpose?
 */

cpl_frame *
xsh_create_master_bias (cpl_frameset* raws, 
                        cpl_frame * frame, 
                        xsh_instrument* instr,
                        cpl_parameterlist* parameters)
{
  xsh_pre *master = NULL;
  cpl_frame *masterFrame = NULL;
  const char* tag = XSH_GET_TAG_FROM_ARM( XSH_MASTER_BIAS, instr);
  char mastername[256];
  int binx=0;
  int biny=0;




  XSH_ASSURE_NOT_NULL(frame);
  check(master = xsh_pre_load (frame,instr));
  /* set the PROCATG KW */
  check(xsh_pfits_set_pcatg (master->data_header, tag));

  check(binx=xsh_pfits_get_binx(master->data_header));
  check(biny=xsh_pfits_get_biny(master->data_header));

  if ( parameters != NULL ) {
     check( set_masterbias_qc_structure(master,parameters));
     if(cpl_frameset_get_size(raws) > 1 ) {
        check(xsh_compute_ron_mbias(raws,master,parameters));
        check(xsh_compute_fpn_mbias(raws,master,instr,parameters));
     }
  }

  /* Saabine doesn't like the bin in the filename
  sprintf(mastername,"MASTER_BIAS_%s_%dx%d.fits", 
			  xsh_instrument_arm_tostring (instr),binx,biny);
  */
  sprintf(mastername,"MASTER_BIAS_%s.fits", 
			  xsh_instrument_arm_tostring (instr));

  xsh_msg ("Create master bias");
  check(masterFrame = xsh_pre_save (master, mastername,tag,1 ));
  cpl_frame_set_tag (masterFrame, tag);
  cpl_frame_set_group(masterFrame,CPL_FRAME_GROUP_CALIB);

  cleanup:
    xsh_pre_free(&master);
    return masterFrame;
}

cpl_frame *
xsh_create_master_flat (cpl_frame *frame, xsh_instrument *instr)
{
  xsh_pre *master = NULL;
  cpl_frame *masterFrame = NULL;
  const char *tag = NULL;
  char *name = NULL;
  double mean = 0.0;
  double median = 0.0;
  double stdev = 0.0;

  /* Check parameters */
  XSH_ASSURE_NOT_NULL( frame);
  XSH_ASSURE_NOT_NULL( instr);

  check( master = xsh_pre_load( frame, instr));

  /* set the PROCATG KW */
  tag = XSH_GET_TAG_FROM_LAMP( XSH_MASTER_FLAT, instr);
  check( xsh_pfits_set_pcatg( master->data_header, tag));
  
  check( xsh_pre_normalize( master));
  /* QC parameters */
  mean = cpl_image_get_mean( master->data);
  median = cpl_image_get_median( master->data);
  stdev = cpl_image_get_stdev( master->data);
  
  //xsh_pfits_set_qc_mbiasavg (master->data_header, mean);
  //xsh_pfits_set_qc_mbiasmed (master->data_header, median);
  //xsh_pfits_set_qc_mbiasrms (master->data_header, stdev);

  XSH_NAME_LAMP_MODE_ARM( name, "MASTER_FLAT",".fits",instr);
  xsh_msg ("Create master flat %s tag %s", name, tag);

  check( masterFrame = xsh_pre_save( master, name, tag,1));
  check( cpl_frame_set_tag( masterFrame, tag));
  check( cpl_frame_set_group( masterFrame, CPL_FRAME_GROUP_CALIB));

  cleanup:
    XSH_FREE( name);
    xsh_pre_free( &master);
    return masterFrame;
}


cpl_frame *
xsh_create_master_bpmap (cpl_frame * mdarkFrame, xsh_instrument* instr)
{
  /*
     Load frame into PRE 
     Create Frame with only the bpmap (qual from xsh_pre)
     Save and returns pointer to frame
  */
  xsh_pre * mdark_pre = NULL ;
  cpl_frame* resFrame = NULL;
  const  char* tag = XSH_GET_TAG_FROM_ARM( XSH_MASTER_BP_MAP, instr);
  char* finalName = NULL;
  int i=0, j=0;
  int* qual_data = NULL;
  int nx,ny;
  int nb_flag_pixels = 0;
  int error_pixels[32];

  /* check parameters input */
  XSH_ASSURE_NOT_NULL( mdarkFrame);
  XSH_ASSURE_NOT_NULL( instr);

  /* load data */
  check( mdark_pre = xsh_pre_load( mdarkFrame, instr));

  if ( xsh_instrument_get_arm(instr) == XSH_ARM_NIR ) {
    check(finalName=cpl_sprintf("MASTER_BP_MAP_%s.fits",
			       xsh_instrument_arm_tostring(instr)));

  } else {
    /* Sabine doesn't like the binning in the filename
    check(binx=xsh_pfits_get_binx(mdark_pre->data_header));
    check(biny=xsh_pfits_get_biny(mdark_pre->data_header));
    check(finalName=cpl_sprintf("MASTER_BP_MAP_%s_%dx%d.fits", 
				xsh_instrument_arm_tostring(instr),binx,biny));
    */
    check(finalName=cpl_sprintf("MASTER_BP_MAP_%s.fits",
			       xsh_instrument_arm_tostring(instr)));

  }

   check(resFrame=xsh_frame_product(finalName,tag,
                                   CPL_FRAME_TYPE_IMAGE,
                                   CPL_FRAME_GROUP_PRODUCT,
                                   CPL_FRAME_LEVEL_FINAL));

  /* Compute some QC KW */
  check( qual_data = cpl_image_get_data_int( mdark_pre->qual));
  check( nx = cpl_image_get_size_x( mdark_pre->qual));
  check( ny = cpl_image_get_size_y( mdark_pre->qual));

  /* count error in BP_MAP */
  for(j=0; j<32; j++){
    error_pixels[j] = 0;
  }

  for(i=0; i<nx*ny; i++){
    if( qual_data[i] > 0){
      nb_flag_pixels++;
      for( j=0;j<32;j++){
        int cmp = pow(2,j);
        if ( qual_data[i] & cmp){
          error_pixels[j]++;
        }
      }
    }
  }


  /* Write QC */
  check( xsh_pfits_set_qc_bp_map_ntotal(mdark_pre->qual_header,nb_flag_pixels));

  for(i=1; i< 33; i++){
    check( xsh_pfits_set_qc_multi( mdark_pre->qual_header, 
      (void *)&(error_pixels[i-1]), XSH_QC_BP_MAP_NFLAGi, instr, i));
  }
  /* Save qual image by itself */
  
  check_msg (cpl_image_save (mdark_pre->qual, finalName, XSH_PRE_DATA_BPP,
			     mdark_pre->qual_header, CPL_IO_DEFAULT),
	     "Could not save bpmap to %s extension", finalName);


  cleanup:
    if (cpl_error_get_code() != CPL_ERROR_NONE){
      xsh_free_frame(&resFrame);
    }
    XSH_FREE(finalName);
    xsh_pre_free( &mdark_pre);
    return resFrame;
}

static cpl_error_code
xsh_compute_ron_mbias(cpl_frameset* raws,
		      xsh_pre* master,
		      cpl_parameterlist* parameters)
{

  const char* ron_method="ALL";
  int random_sizex=-1;
  int random_nsamples=-1;
  double ron=0;
  double ron_err=0;


  int ref_llx=-1;
  int ref_lly=-1;
  int ref_urx=-1;
  int ref_ury=-1;


  int ref2_llx=-1;
  int ref2_lly=-1;
  int ref2_urx=-1;
  int ref2_ury=-1;


  cpl_parameter* p=NULL;

  const cpl_frame  * fr = NULL;

  cpl_propertylist * plist = NULL;

  int naxis1 = 0;
  int naxis2 = 0;

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.ron_method"));
  check(ron_method = cpl_parameter_get_string(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.random_sizex"));
  check(random_sizex = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.random_nsamples"));
  check(random_nsamples = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.ref1_llx"));
  check(ref_llx = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.ref1_lly"));
  check(ref_lly = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.ref1_urx"));
  check(ref_urx = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.ref1_urx"));
  check(ref_ury = cpl_parameter_get_int(p));

  check(fr=cpl_frameset_get_first_const(raws));

  check(plist=cpl_propertylist_load(cpl_frame_get_filename(fr),0));
  check(naxis1 = cpl_propertylist_get_int(plist, "NAXIS1"));
  check(naxis2 = cpl_propertylist_get_int(plist, "NAXIS2"));


  cpl_propertylist_delete(plist);

  //checking for defaults
  if(ref_llx == -1) ref_llx = naxis1 / 8;
  if(ref_lly == -1) ref_lly = naxis2 / 8;
  if(ref_urx == -1) ref_urx = naxis1 * 7 / 8;
  if(ref_ury == -1) ref_ury = naxis2 * 7 / 8;


  check(xsh_compute_ron(raws,ref_llx,ref_lly,ref_urx,ref_ury,random_nsamples,random_sizex,1,&ron,&ron_err));
  xsh_pfits_set_qc_ron1(master->data_header,ron);
  xsh_pfits_set_qc_ron1_err(master->data_header,ron_err);


  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.ref2_llx"));
  check(ref2_llx = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.ref2_lly"));
  check(ref2_lly = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.ref2_urx"));
  check(ref2_urx = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.ref2_urx"));
  check(ref2_ury = cpl_parameter_get_int(p));


  //checking for defaults
  if(ref2_llx == -1) ref2_llx = naxis1 / 8;
  if(ref2_lly == -1) ref2_lly = naxis2 / 8;
  if(ref2_urx == -1) ref2_urx = naxis1 * 7 / 8;
  if(ref2_ury == -1) ref2_ury = naxis2 * 7 / 8;

  if( 
     (ref2_llx != ref_llx) ||
     (ref2_lly != ref_lly) ||
     (ref2_urx != ref_urx) ||
     (ref2_ury != ref_ury) 
     ) {

  check(xsh_compute_ron(raws,ref2_llx,ref2_lly,ref2_urx,ref2_ury,random_nsamples,random_sizex,2,&ron,&ron_err));

  xsh_pfits_set_qc_ron2(master->data_header,ron);
  xsh_pfits_set_qc_ron2_err(master->data_header,ron_err);

  }


 cleanup:

  return cpl_error_get_code();

}


static cpl_error_code
xsh_compute_fpn_mbias(cpl_frameset* raws,
		      xsh_pre* master,
		      xsh_instrument* instrument,
		      cpl_parameterlist* parameters)
{

    int nraws=0;
    double ron=0;
    double fpn=0;
    cpl_frame* frm1=NULL;
    cpl_frame* frm2=NULL;
    cpl_image* ima1=NULL;
    cpl_image* ima2=NULL;
    cpl_image* ima=NULL;


 if ( xsh_instrument_get_arm(instrument) != XSH_ARM_NIR){
    nraws=cpl_frameset_get_size(raws);
    if(nraws > 1) {
       frm1=cpl_frameset_get_first(raws);
       frm2=cpl_frameset_get_next(raws);
       ima1=cpl_image_load(cpl_frame_get_filename(frm1),CPL_TYPE_FLOAT,0,0);
       ima2=cpl_image_load(cpl_frame_get_filename(frm2),CPL_TYPE_FLOAT,0,0);

       ima = cpl_image_duplicate(ima1);
       check(ron = xsh_image_get_stdev_clean(ima, NULL) / sqrt(2.0));

       check(fpn = xsh_fixed_pattern_noise_bias(ima1,ima2,ron));

       xsh_pfits_set_qc_ron_master(master->data_header,ron);
       xsh_pfits_set_qc_fpn_master(master->data_header,fpn);

       xsh_free_image(&ima);
       xsh_free_image(&ima1);
       xsh_free_image(&ima2);
    }
 } else {
    check(xsh_mbias_get_fpn(master->data,parameters,&fpn));
    xsh_msg("Fixed Pattern Noise=%f",fpn);

 }

 cleanup:
 xsh_free_image(&ima);
 xsh_free_image(&ima1);
 xsh_free_image(&ima2);
 return cpl_error_get_code();

}


/*---------------------------------------------------------------------------*/

/*
 * @brief  Retrieve input parameters
 * @param  pipeline_name        Input image
 * @param  recipe_name          Input image
 * @param  parlist              Shift to apply on the x-axis
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static cpl_error_code
xsh_my_detmon_ronbias_random(const cpl_imagelist * rawbiases,
			     const cpl_image     * masterbias,
                             cpl_propertylist    * qclist,
                             const char* method,
                             const int random_sizex,
                             const int random_nsamples)
{
    int                     nraws = cpl_imagelist_get_size(rawbiases);
    int                     i;
    double                  bias = DBL_MAX;  /* Avoid (false) uninit warning */
    double                  bias_error;

    double                  ron_error;
    double                 *ron =
        (double *) cpl_malloc(sizeof(double) * nraws);

    cpl_vector             *v;
    double stdev;
    cpl_error_code error;

    /* As we are applying to diff frames instead of raw frames,
       there is one less to compute on */
    if(!strcmp(method, "DIF")) {
	nraws--;
	/* As we are applying to diff frames instead of raw frames,
	   there is one less to compute on */
	for(i = 0; i < nraws; i++) {
	    const cpl_image        *c1_raw =
		cpl_imagelist_get_const(rawbiases, i);
	    const cpl_image        *c2_raw =
		cpl_imagelist_get_const(rawbiases, i + 1);
	    /*FIXME: See if const modifier is necessary */
	    const cpl_image        *c_raw = cpl_image_subtract_create(c1_raw,
								      c2_raw);
	    cpl_flux_get_noise_window(c_raw, NULL,
				      random_sizex / 2,
				      random_nsamples,
				      ron + i, &ron_error);
	    cpl_image_delete((cpl_image*)c_raw);
	}

    } else {
	for(i = 0; i < nraws; i++) {
	    const cpl_image        *c_raw = cpl_imagelist_get_const(rawbiases, i);
	    cpl_flux_get_noise_window(c_raw, NULL,
				      random_sizex / 2,
				      random_nsamples,
				      ron + i, &ron_error);
	}
    }


    /*FIXME: Calls to noise_window could be out from if() and
      coded only once */

    xsh_flux_get_bias_window(masterbias, NULL,
				random_sizex / 2,
				random_nsamples,
				&bias, &bias_error);

    v = cpl_vector_wrap(nraws, ron);
    stdev = cpl_vector_get_median_const(v);
    cpl_vector_unwrap(v);

    cpl_free(ron);

    error = cpl_propertylist_append_double(qclist, "ESO QC BIAS RANDOM VAL", bias);
    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist, "ESO QC BIAS RANDOM RON", stdev);    //ron_error); //stdev);
    cpl_ensure_code(!error, error);
    return cpl_error_get_code();
}




static cpl_error_code
xsh_my_detmon_ronbias_histo(const cpl_imagelist * rawbiases,
			 const cpl_image * masterbias,
			 cpl_propertylist * qclist,
                         const char* method)
{
    int                     nraws = cpl_imagelist_get_size(rawbiases);
    int                     i;

    double mbias = DBL_MAX;
    double mfwhm = DBL_MAX;
    double mmax = DBL_MAX;

    cpl_vector * fwhms;
    cpl_vector * maxs;

    double mean_fwhm = DBL_MAX;

    if(!strcmp(method, "DIF")) nraws--;

    fwhms = cpl_vector_new(nraws);
    maxs = cpl_vector_new(nraws);

    for(i = 0; i < nraws; i++) {
	/*FIXME: See if it is necessary to have const */
	const cpl_image * c_raw;
	double bias = DBL_MAX;
	double fwhm = DBL_MAX;
	double max = DBL_MAX;

	if(strcmp(method, "DIF")) {
           c_raw = cpl_imagelist_get_const(rawbiases, i);
	} else {
	    const cpl_image        *c1_raw = cpl_imagelist_get_const(rawbiases, i);
	    const cpl_image        * c2_raw = cpl_imagelist_get_const(rawbiases, i+1);
	    c_raw = cpl_image_subtract_create(c1_raw, c2_raw);
	}

	skip_if(xsh_detmon_ronbias_histo_reduce(c_raw, &bias, &fwhm, &max));

	skip_if(bias == DBL_MAX || fwhm == DBL_MAX || max == DBL_MAX);

	if(!strcmp(method, "DIF"))
	    cpl_image_delete((cpl_image *)c_raw);

        skip_if(cpl_vector_set(maxs, i, max));
        skip_if(cpl_vector_set(fwhms, i, fwhm));

        //cpl_msg_error(cpl_func, "%d", xsh_hist_get_size(hist));
//      char *       name = cpl_sprintf("hist_pr%d.fits", i);
	//      cpl_table * table = xsh_hist_cast_table(hist);
	//cpl_table_save(table, NULL, NULL, name, CPL_IO_DEFAULT);
//      cpl_table_delete(table);
    }

    skip_if(cpl_vector_divide_scalar(fwhms, HIST_FACT));

    xsh_detmon_ronbias_histo_reduce(masterbias, &mbias, &mfwhm, &mmax);

    cpl_msg_warning(cpl_func,"mfwhm=%f",mfwhm);
    skip_if(mbias == DBL_MAX || mfwhm == DBL_MAX || mmax == DBL_MAX);

    skip_if(cpl_propertylist_append_double(qclist, "ESO QC BIAS HISTO VAL",
                                           mbias));
    mean_fwhm = cpl_vector_get_mean(fwhms);

    skip_if(mean_fwhm == DBL_MAX);
    skip_if(cpl_error_get_code());

    skip_if(cpl_propertylist_append_double(qclist, "ESO QC BIAS HISTO RON",
					   mean_fwhm));

    //  skip_if(cpl_propertylist_append_double(qclist,
//    				   "ESO QC BIAS HISTO RMS STDEV",
					   //			   cpl_vector_get_stdev(fwhms)));
//    skip_if(cpl_propertylist_append_double(qclist, "ESO QC BIAS HISTO MAX",
    //                                         cpl_vector_get_mean(maxs)));
//    skip_if(cpl_propertylist_append_double(qclist,
    //				   "ESO QC BIAS HISTO MAX STDEV",
    //				   cpl_vector_get_stdev(maxs)));

    end_skip;

    cpl_vector_delete(fwhms);
    cpl_vector_delete(maxs);

    return cpl_error_get_code();
}


/*---------------------------------------------------------------------------*/

/*
 * @brief  Retrieve input parameters
 * @param  pipeline_name        Input image
 * @param  recipe_name          Input image
 * @param  parlist              Shift to apply on the x-axis
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static cpl_error_code
xsh_my_detmon_ronbias_region(const cpl_imagelist * rawbiases,
			     const cpl_image * masterbias,
                             cpl_propertylist * qclist,
                             const char* pmethod,
                             const int ref_llx,
                             const int ref_lly,
                             const int ref_urx,
                             const int ref_ury,
                             const int stacking_ks_low,
			     const int stacking_ks_iter)
{

    int                     nraws = cpl_imagelist_get_size(rawbiases);
    int                     i;

    int                     nx =
        cpl_image_get_size_x(cpl_imagelist_get_const(rawbiases, 0));
    int                     ny =
        cpl_image_get_size_y(cpl_imagelist_get_const(rawbiases, 0));

    cpl_vector *rmssreg;
    cpl_error_code          error;

    const cpl_image * c_raw;
    double median, mbias, mstdev;

    if(!strcmp(pmethod, "DIF")) nraws--;

    rmssreg = cpl_vector_new(nraws);

    if(nx < ref_urx ||
       ny < ref_ury) {
        cpl_msg_warning(cpl_func, "REGION method not applied. Given "
                        "limits of prescan and overscan area "
                        "exceed image size. Please check and rerun.");
        return CPL_ERROR_NONE;
    }

    for(i = 0; i < nraws; i++) {
        double                  mean = 0;
        double                  stdev = 0;
	if(strcmp(pmethod, "DIF")) {
	    c_raw = cpl_imagelist_get_const(rawbiases, i);
	} else {
	    const cpl_image        *c1_raw = cpl_imagelist_get_const(rawbiases, i);
	    const cpl_image        * c2_raw = cpl_imagelist_get_const(rawbiases, i+1);
	    c_raw = cpl_image_subtract_create(c1_raw, c2_raw);
	}
        error = xsh_ksigma_clip(c_raw,
                                             ref_llx,
                                             ref_lly,
                                             ref_urx,
                                             ref_ury,
                                             (double) stacking_ks_low,
                                             stacking_ks_iter, 1e-5,
					     &mean, &stdev);

        cpl_ensure_code(!error, error);
        /*        cpl_vector_set(rmssreg, i, cpl_image_get_stdev_window(c_raw,
           detmon_ronbias_config.ref_llx,
           detmon_ronbias_config.ref_lly,
           detmon_ronbias_config.ref_urx,    
           detmon_ronbias_config.ref_ury));
         */
        error = cpl_vector_set(rmssreg, i, stdev);
        cpl_ensure_code(!error, error);
	if(!strcmp(pmethod, "DIF")) cpl_image_delete((cpl_image *)c_raw);
    }

    median =  cpl_image_get_median_window(masterbias,
					  ref_llx,
					  ref_lly,
					  ref_urx,
					  ref_ury);
    error = xsh_ksigma_clip(masterbias,
					 ref_llx,
					 ref_lly,
					 ref_urx,
					 ref_ury,
					 (double) stacking_ks_low,
					 stacking_ks_iter, 1e-5,
					 &mbias, &mstdev);


    error = cpl_propertylist_append_double(qclist, "ESO QC BIAS REGION MED",
                                           median);
    cpl_ensure_code(!error, error);

    error = cpl_propertylist_append_double(qclist, "ESO QC BIAS REGION VAL",
                                           mbias);
    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist, "ESO QC BIAS REGION RON",
                                           cpl_vector_get_mean(rmssreg));
    cpl_ensure_code(!error, error);
    /*
    error =
        cpl_propertylist_append_double(qclist, "ESO QC BIAS REGION RMS STDEV",
                                       cpl_vector_get_stdev(rmssreg));
    cpl_ensure_code(!error, error);
    */
    cpl_vector_delete(rmssreg);

    return cpl_error_get_code();
}

static cpl_error_code
xsh_mbias_get_fpn(const cpl_image* ima,
                  cpl_parameterlist* parameters,
                  double* fpn)
{

  int fpn_llx=0;
  int fpn_lly=0;
  int fpn_urx=0;
  int fpn_ury=0;
  int fpn_hsize=0;
  int fpn_nsamp=0;
  cpl_parameter* p=NULL;
 
  int zone[4];
  int sx=0;
  int sy=0;

   check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.fpn_llx"));
  check(fpn_llx = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.fpn_lly"));
  check(fpn_lly = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.fpn_urx"));
  check(fpn_urx = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.fpn_ury"));
  check(fpn_ury = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.fpn_hsize"));
  check(fpn_hsize = cpl_parameter_get_int(p));

  check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mbias.fpn_nsamples"));
  check(fpn_nsamp = cpl_parameter_get_int(p));

  sx=cpl_image_get_size_x(ima);
  sy=cpl_image_get_size_y(ima);

  fpn_llx=(fpn_llx>0) ? fpn_llx: 0;
  fpn_lly=(fpn_lly>0) ? fpn_lly: 0;

  fpn_urx=(fpn_urx<sx) ? fpn_urx: sx;
  fpn_ury=(fpn_ury<sy) ? fpn_ury: sy;

  
  zone[0]=fpn_llx;
  zone[1]=fpn_urx;
  zone[2]=fpn_lly;
  zone[3]=fpn_ury;

  
  check_msg(cpl_flux_get_noise_window(ima,zone,fpn_hsize,fpn_nsamp,fpn,NULL),
         "Error computing noise in a window");

 cleanup:
  return cpl_error_get_code();

}



/*---------------------------------------------------------------------------*/

/*
 * @brief  Retrieve input parameters
 * @param  pipeline_name        Input image
 * @param  recipe_name          Input image
 * @param  parlist              Shift to apply on the x-axis
 * @return CPL_ERROR_NONE on success.
 */

/*---------------------------------------------------------------------------*/
static cpl_error_code
xsh_my_detmon_ronbias_preoverscan(const cpl_imagelist * rawbiases,
//				  const cpl_image * masterbias,
                                  cpl_propertylist * qclist,
                                  cpl_image ** synthetic,
                                  const int prescan_llx,
                                  const int prescan_lly,
                                  const int prescan_urx,
                                  const int prescan_ury,
                                  const int overscan_llx,
                                  const int overscan_lly,
                                  const int overscan_urx,
                                  const int overscan_ury,
                                  const int stacking_ks_low,
                                  const int stacking_ks_iter)
{
    int                     i;
    int                     nx, ny;
    int                     nraws;

    cpl_vector             *meanspre;
    cpl_vector             *medspre;
    cpl_vector             *rmsspre;
    cpl_vector             *meansover;
    cpl_vector             *medsover;
    cpl_vector             *rmssover;

    cpl_error_code          error;

    nraws = cpl_imagelist_get_size(rawbiases);
    cpl_ensure_code(nraws != -1, CPL_ERROR_ILLEGAL_INPUT);

    meanspre = cpl_vector_new(nraws);
    medspre = cpl_vector_new(nraws);
    rmsspre = cpl_vector_new(nraws);
    meansover = cpl_vector_new(nraws);
    medsover = cpl_vector_new(nraws);
    rmssover = cpl_vector_new(nraws);

    nx = cpl_image_get_size_x(cpl_imagelist_get_const(rawbiases, 0));
    ny = cpl_image_get_size_y(cpl_imagelist_get_const(rawbiases, 0));
    cpl_ensure_code(nx != -1 && ny != -1, CPL_ERROR_ILLEGAL_INPUT);

    if(nx < prescan_urx ||
       nx < overscan_urx ||
       ny < prescan_ury ||
       ny < overscan_ury) {
        cpl_msg_warning(cpl_func, "PREOVERSCAN method not applied. Given "
                        "limits of prescan and overscan area "
                        "exceed image size. Please check and rerun.");
        return CPL_ERROR_NONE;
    }

    for(i = 0; i < nraws; i++) {
        double                  mean = 0;
        double                  stdev = 0;

        cpl_image              *prescan = NULL;
        cpl_image              *overscan = NULL;

        const cpl_image        *c_raw = cpl_imagelist_get_const(rawbiases, i);

        cpl_ensure_code(c_raw != NULL, CPL_ERROR_ILLEGAL_INPUT);

        prescan =
            cpl_image_extract(c_raw,
                              prescan_llx,
                              prescan_lly,
                              prescan_urx,
                              prescan_ury);
        cpl_ensure_code(prescan != NULL, CPL_ERROR_ILLEGAL_INPUT);
        overscan =
            cpl_image_extract(c_raw,
                              overscan_llx,
                              overscan_lly,
                              overscan_urx,
                              overscan_ury);
        cpl_ensure_code(overscan != NULL, CPL_ERROR_ILLEGAL_INPUT);

        if(i == 0) {
            *synthetic = xsh_detmon_build_synthetic(prescan, overscan);
            cpl_msg_info(cpl_func, "Creating SYNTHETIC frame");
            if(*synthetic == NULL) {
                cpl_msg_error(cpl_func, "Error creating SYNTHETIC frame");
                return CPL_ERROR_UNSPECIFIED;
            }
        }

        error = xsh_ksigma_clip(c_raw,prescan_llx,prescan_lly,
                                prescan_urx,prescan_ury,
                                (double) stacking_ks_low,stacking_ks_iter, 
				1e-5,&mean, &stdev);
        cpl_ensure_code(!error, error);

        cpl_ensure_code(mean != 0 && stdev != 0, CPL_ERROR_UNSPECIFIED);

        error = cpl_vector_set(medspre, i, cpl_image_get_median(prescan));
        cpl_ensure_code(!error, error);
        //        cpl_vector_set(meanspre, i, cpl_image_get_mean(prescan));
        //        cpl_vector_set(rmsspre,  i, cpl_image_get_stdev(prescan));
        error = cpl_vector_set(meanspre, i, mean);
        cpl_ensure_code(!error, error);
        error = cpl_vector_set(rmsspre, i, stdev);
        cpl_ensure_code(!error, error);
        error = xsh_ksigma_clip(c_raw,overscan_llx,overscan_lly,
				overscan_urx,overscan_ury,
				(double) stacking_ks_low,stacking_ks_iter, 
				1e-5,&mean, &stdev);

        cpl_ensure_code(!error, error);

        cpl_ensure_code(mean != 0 && stdev != 0, CPL_ERROR_UNSPECIFIED);

        error = cpl_vector_set(medsover, i, cpl_image_get_median(overscan));
        cpl_ensure_code(!error, error);

        //        cpl_vector_set(meansover, i, cpl_image_get_mean(overscan));
        //        cpl_vector_set(rmssover,  i, cpl_image_get_stdev(overscan));
        error = cpl_vector_set(meansover, i, mean);
        cpl_ensure_code(!error, error);
        error = cpl_vector_set(rmssover, i, stdev);
        cpl_ensure_code(!error, error);

        cpl_image_delete(prescan);
        cpl_image_delete(overscan);
    }

    error = cpl_propertylist_append_double(qclist, "ESO QC BIAS PRESCAN MEAN",
                                           cpl_vector_get_mean(meanspre));
    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist, "ESO QC BIAS PRESCAN MED",
                                           cpl_vector_get_mean(medspre));
    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist, "ESO QC BIAS PRESCAN RON",
                                           cpl_vector_get_mean(rmsspre));
    cpl_ensure_code(!error, error);

    error =
        cpl_propertylist_append_double(qclist, "ESO QC BIAS OVERSCAN MEAN",
                                       cpl_vector_get_mean(meansover));
    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist, "ESO QC BIAS OVERSCAN MED",
                                           cpl_vector_get_mean(medsover));
    cpl_ensure_code(!error, error);
    error = cpl_propertylist_append_double(qclist, "ESO QC BIAS OVERSCAN RON",
                                           cpl_vector_get_mean(rmssover));
    cpl_ensure_code(!error, error);

    /* The following seems not to be necessary.
       Pending of revision to be removed */
    /*
    error =
        cpl_propertylist_append_double(qclist,
                                       "ESO QC BIAS PRESCAN MEAN STDEV",
                                       cpl_vector_get_stdev(meanspre));
    cpl_ensure_code(!error, error);
    error =
        cpl_propertylist_append_double(qclist,
                                       "ESO QC BIAS PRESCAN MED STDEV",
                                       cpl_vector_get_stdev(medspre));
    cpl_ensure_code(!error, error);
    error =
        cpl_propertylist_append_double(qclist,
                                       "ESO QC BIAS PRESCAN RMS STDEV",
                                       cpl_vector_get_stdev(rmsspre));
    cpl_ensure_code(!error, error);

    error =
        cpl_propertylist_append_double(qclist,
                                       "ESO QC BIAS OVERSCAN MEAN STDEV",
                                       cpl_vector_get_stdev(meansover));
    cpl_ensure_code(!error, error);
    error =
        cpl_propertylist_append_double(qclist,
                                       "ESO QC BIAS OVERSCAN MED STDEV",
                                       cpl_vector_get_stdev(medsover));
    cpl_ensure_code(!error, error);
    error =
        cpl_propertylist_append_double(qclist,
                                       "ESO QC BIAS OVERSCAN RMS STDEV",
                                       cpl_vector_get_stdev(rmssover));
    cpl_ensure_code(!error, error);
    */

    cpl_vector_delete(meanspre);
    cpl_vector_delete(medspre);
    cpl_vector_delete(rmsspre);
    cpl_vector_delete(meansover);
    cpl_vector_delete(medsover);
    cpl_vector_delete(rmssover);

    return CPL_ERROR_NONE;
}


/**@}*/
