/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $
 * $Date: 2011/03/09 14:38:12 $
 * $Revision: 1.14 $
 *
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------
                                 Includes
 ----------------------------------------------------------------------------*/
#include <xsh_utils_scired_slit.h>
#include <xsh_dfs.h>
#include <xsh_utils.h>
#include <xsh_error.h>
#include <xsh_drl_check.h>
#include <xsh_pfits.h>
#include <xsh_model_utils.h>
static int 
xsh_frame_has_same_nod_pos(cpl_frame* ref, cpl_frame* cmp)
{
  int result=0;

  cpl_propertylist* href=NULL;
  cpl_propertylist* hcmp=NULL;

  const char* name_ref=NULL;
  const char* name_cmp=NULL;

  double ref_yoff=0;
  double ref_ra=0;
  double ref_dec=0;

  double cmp_yoff=0;
  double cmp_ra=0;
  double cmp_dec=0;

  name_ref=cpl_frame_get_filename(ref);
  name_cmp=cpl_frame_get_filename(cmp);
  href=cpl_propertylist_load(name_ref,0);
  hcmp=cpl_propertylist_load(name_cmp,0);

  if(cpl_propertylist_has(href, XSH_NOD_CUMULATIVE_OFFSETY)){

     check(ref_yoff=xsh_pfits_get_cumoffsety(href));
     check(cmp_yoff=xsh_pfits_get_cumoffsety(hcmp));

   if(ref_yoff==cmp_yoff) {
     result=1;
   } 

  } else {

     check(ref_ra=xsh_pfits_get_ra_cumoffset(href));
     check(ref_dec=xsh_pfits_get_dec_cumoffset(href));

     check(cmp_ra=xsh_pfits_get_ra_cumoffset(hcmp));
     check(cmp_dec=xsh_pfits_get_dec_cumoffset(hcmp));

    if((ref_ra==cmp_ra) && (ref_ra==cmp_ra)) {
     result=1;
    } 

  }
  cleanup:
  xsh_free_propertylist(&href);
  xsh_free_propertylist(&hcmp);

  return result;

}

cpl_frameset* xsh_nod_group_by_reloff( cpl_frameset *ord_set,
  xsh_instrument *instrument, xsh_clipping_param *crh_clipping_par)
{

  cpl_frameset *result = NULL;
  cpl_frameset *rmcrh_set = NULL;
  cpl_frame *frame = NULL;
  cpl_propertylist *header = NULL;

  const char* nod_name = NULL;

  double rel_ra=0;
  double rel_dec=0;
  double A_cum_ra=0;
  double A_cum_dec=0;
  int A_number=1, B_number =1;
  int nod_number;
  char ftag[40];
  cpl_frame *crhm_frame = NULL;

  XSH_ASSURE_NOT_NULL( ord_set);
  XSH_ASSURE_NOT_NULL( instrument);
  XSH_ASSURE_NOT_NULL( crh_clipping_par);

  check( result=cpl_frameset_new());

  check( frame = cpl_frameset_get_first( ord_set));
  check( nod_name = cpl_frame_get_filename( frame));  
  check( header = cpl_propertylist_load( nod_name, 0));

  check( A_cum_ra = xsh_pfits_get_ra_cumoffset( header));
  check( A_cum_dec = xsh_pfits_get_dec_cumoffset( header));

  xsh_free_propertylist( &header);

  while( frame != NULL){
    double cum_ra=0;
    double cum_dec=0;
    double nrel_ra, nrel_dec;
    const char *type= NULL;

    check( nod_name = cpl_frame_get_filename( frame));
    check( header = cpl_propertylist_load( nod_name, 0));
    check( cum_ra = xsh_pfits_get_ra_cumoffset( header));
    check( cum_dec = xsh_pfits_get_dec_cumoffset( header));    
    check( rel_ra = xsh_pfits_get_ra_reloffset( header));
    check( rel_dec = xsh_pfits_get_dec_reloffset( header));

    if ( cum_ra == A_cum_ra && cum_dec == A_cum_dec){
      type = "A";
      nod_number = A_number;
    }
    else{
      type = "B";
      nod_number = B_number;
    }
    
    xsh_msg("name %s cumoffset %f %f reloffset %f %f nod_seq_number : %s%d", nod_name,
      cum_ra, cum_dec, rel_ra, rel_dec, type, nod_number);

    xsh_free_propertylist( &header);
 
    check( rmcrh_set = cpl_frameset_new());
    cpl_frameset_insert( rmcrh_set, cpl_frame_duplicate( frame));

    check( frame = cpl_frameset_get_next( ord_set));

    while( frame != NULL){
      check( nod_name = cpl_frame_get_filename( frame));
      check( header = cpl_propertylist_load( nod_name, 0));
      check( nrel_ra = xsh_pfits_get_ra_reloffset( header));
      check( nrel_dec = xsh_pfits_get_dec_reloffset( header));
      xsh_free_propertylist( &header);

      if ( rel_ra == nrel_ra && nrel_dec == rel_dec){
        cpl_frameset_insert( rmcrh_set, cpl_frame_duplicate( frame));
      }
      else{
        break;
      }
      check( frame = cpl_frameset_get_next( ord_set));
    }

    /* Create the remove cosmics frame */
    sprintf( ftag,"med_%s%d", type, nod_number);
    check( crhm_frame = xsh_remove_crh_multiple( rmcrh_set, ftag, crh_clipping_par,
                                                 instrument, NULL,NULL ));

    check( cpl_frameset_insert( result, crhm_frame));

    if ( cum_ra == A_cum_ra && cum_dec == A_cum_dec){
      A_number++;
    }
    else{
      B_number++;
    }
    xsh_free_frameset( &rmcrh_set);
  }

  cleanup:
    if ( cpl_error_get_code() != CPL_ERROR_NONE){
      xsh_free_propertylist( &header);
      xsh_free_frameset( &rmcrh_set);
      xsh_free_frameset( &result);
    }
    return result;
}

static cpl_frameset* 
xsh_nod_prepare_set( cpl_frameset *ord_set,
		     xsh_instrument* instrument,
		     xsh_clipping_param* crh_clipping_par)
{

  int nraw=0;
  int i=0;
  int k=0;

  cpl_frameset* result=NULL;
  cpl_frameset* set_tmp=NULL;
  cpl_frame* frm=NULL;
  cpl_frame* frm_tmp=NULL;
  cpl_frame* crhm_frame=NULL;
  char ftag[40];

  result=cpl_frameset_new();
  frm=cpl_frameset_get_frame(ord_set,0);
  set_tmp=cpl_frameset_new();
  cpl_frameset_insert(set_tmp,cpl_frame_duplicate(frm));

  nraw=cpl_frameset_get_size( ord_set);  
  /* xsh_msg("nraw=%d",nraw); */
  for(i=1;i<nraw;i++) {

    frm_tmp=cpl_frameset_get_frame(ord_set,i);

    if(xsh_frame_has_same_nod_pos(frm,frm_tmp)) {
      cpl_frameset_insert(set_tmp,cpl_frame_duplicate(frm_tmp));
  
    } else {
      sprintf(ftag,"crh_mean_%d",k);
      check(crhm_frame = xsh_remove_crh_multiple(set_tmp,
						 ftag,
						 crh_clipping_par,
						 instrument, NULL,NULL ));
      xsh_free_frameset(&set_tmp);
      cpl_frameset_insert(result,crhm_frame);
      set_tmp=cpl_frameset_new();
      frm=cpl_frameset_get_frame(ord_set,i);
      cpl_frameset_insert(set_tmp,cpl_frame_duplicate(frm));
      k++;


    }    

  }
  sprintf(ftag,"crh_mean_%d",k);
  check(crhm_frame = xsh_remove_crh_multiple(set_tmp,
					     ftag,
					     crh_clipping_par,
					     instrument, NULL,NULL ));

  cpl_frameset_insert(result,crhm_frame);

  cleanup:
    xsh_free_frameset(&set_tmp);
    return result;

}
/*---------------------------------------------------------------------------*/
/**
  @internal
  @brief    Get the offsets
  @param    rawframes   the input rawframes
  @return   the offsets in a vector
 */
/*---------------------------------------------------------------------------*/
static cpl_vector * 
xsh_get_offsets(const cpl_frameset * rawframes)
{
    int          nraw = 0;
    cpl_vector       * offsets = NULL;
    cpl_propertylist * plist = NULL;
    int                i;

    double ra_off=0;
    double dec_off=0;
    double posang=0;

    XSH_ASSURE_NOT_NULL_MSG (rawframes,"null input frameset");
     
    nraw = cpl_frameset_get_size(rawframes);
    XSH_ASSURE_NOT_ILLEGAL_MSG(nraw > 0,"Provide at least a raw frame");
 
    offsets = cpl_vector_new(nraw);

    /* Get the rawframes Y offsets */
    for (i=0; i < nraw; i++) {
      const cpl_frame * cur_frame=NULL;
      const char * filename = NULL;
      double yoff;
      check(cur_frame= cpl_frameset_get_frame_const(rawframes, i));
      check(filename=cpl_frame_get_filename(cur_frame));

      xsh_free_propertylist(&plist);
      plist=cpl_propertylist_load(filename,0);
      if(cpl_propertylist_has(plist,XSH_NOD_CUMULATIVE_OFFSETY)) {
         check(yoff=xsh_pfits_get_cumoffsety(plist));
      } else {
         ra_off=xsh_pfits_get_ra_cumoffset(plist);
         dec_off=xsh_pfits_get_dec_cumoffset(plist);
         posang = xsh_pfits_get_posang(plist);
         posang = posang/180.0* M_PI;
/*
         shift_pix = xsh_round_double((cos(-posang)*dec_off+
                                       sin(-posang)*ra_off)/bin_space);
*/
         yoff=(cos(-posang)*dec_off+
                  sin(-posang)*ra_off);
         xsh_msg_dbg_medium("ra_off=%f,dec_off=%f,yoff=%f,posang=%f",
                 ra_off,dec_off,yoff,posang);

      }


      check((cpl_vector_set(offsets, i, yoff)));
    }

  cleanup:

    xsh_free_propertylist(&plist);

    if (cpl_error_get_code()) {
        xsh_free_vector(&offsets);
        offsets = NULL;
    }

    return offsets;
}

static
cpl_vector* xsh_get_noffsets(cpl_vector * offsets)
{

    int size=0;
    cpl_vector* result=NULL;
    cpl_table* tv=NULL;
    int nrow=0;
    double* values=NULL;
    double val=0;
    int null=0;
    int noffsets=0;
    XSH_ASSURE_NOT_NULL_MSG (offsets,"null input offsets");

    size=cpl_vector_get_size(offsets);
    values=cpl_vector_get_data(offsets);
    tv=cpl_table_new(size);
    cpl_table_wrap_double(tv,values,"VALUES");

    nrow=cpl_table_get_nrow(tv);
    result=cpl_vector_new(nrow);
    while(nrow>0){
      val=cpl_table_get_double(tv,"VALUES",0,&null);
      cpl_table_and_selected_double(tv,"VALUES",CPL_EQUAL_TO,val);
      cpl_table_erase_selected(tv);
      nrow=cpl_table_get_nrow(tv);
       cpl_vector_set(result,noffsets,val);
      noffsets++;
    }
    cpl_vector_set_size(result,noffsets);
  cleanup:
    xsh_free_table(&tv);

    return result;
}

/* Not used
static cpl_frameset*
xsh_avg_raws_same_off(cpl_frameset* raws,
			cpl_vector* offsets,
			xsh_clipping_param* crh_clipping_par,
			xsh_instrument* instrument)
{

  int ngroups=cpl_vector_get_size(offsets);

  cpl_frame* frm=NULL;
  cpl_frame* crhm_frame=NULL;
  cpl_frameset* set_tmp=0;
  cpl_frameset* result=NULL;
  cpl_propertylist* plist=NULL;

  const char* name=NULL;
  int i=0;
  int j=0;
  double val=0;
  double offy=0;
  char ftag[80];
  int nraw=0;


   double ra_off=0;
   double dec_off=0;
   double posang=0;

  XSH_ASSURE_NOT_NULL_MSG (raws,"null input raws");
  XSH_ASSURE_NOT_NULL_MSG (offsets,"null input offsets");

  nraw=cpl_frameset_get_size(raws);
  result=cpl_frameset_new();
 
  for(i=0;i<ngroups;i++) {
    val=cpl_vector_get(offsets,i);
    set_tmp=cpl_frameset_new();
    for(j=0;j<nraw;j++) {
      frm=cpl_frameset_get_frame(raws,j);
      name=cpl_frame_get_filename(frm);
      plist=cpl_propertylist_load(name,0);
      if(cpl_propertylist_has(plist,XSH_NOD_CUMULATIVE_OFFSETY)) {
         check(offy=xsh_pfits_get_cumoffsety(plist));
      } else {
         check(ra_off=xsh_pfits_get_ra_cumoffset(plist));
         check(dec_off=xsh_pfits_get_dec_cumoffset(plist));
         check(posang = xsh_pfits_get_posang(plist));
         posang = posang/180.0* M_PI;

         offy=(cos(-posang)*dec_off+
                  sin(-posang)*ra_off);
         xsh_msg_dbg_medium("ra_off=%f,dec_off=%f,offy=%f,posang=%f",
                 ra_off,dec_off,offy,posang);
      }
      if(offy==val) {
	cpl_frameset_insert(set_tmp,frm);
      }
    }
  //xsh_frameset_dump(set_tmp);
    sprintf(ftag,"crh_mean_%d",i);
    check(crhm_frame = xsh_remove_crh_multiple(set_tmp,
					       ftag,
					       crh_clipping_par,
					       instrument, NULL,NULL ));
    cpl_frameset_insert(result,cpl_frame_duplicate(crhm_frame));
    //xsh_free_frameset(&set_tmp);
  }

 cleanup:

    return result;

}    
*/

static cpl_frame* xsh_frame_divide_flat( int do_flat, cpl_frame* src, 
  cpl_frame *mflat, const char* tag, xsh_instrument* instr)
{
  cpl_frame *result = NULL;

  if ( do_flat == CPL_TRUE){
    xsh_msg( "Divide by flat");
    check( result = xsh_divide_flat( src, mflat, tag, instr));
  }
  else{
    check( result = cpl_frame_duplicate( src));
  }

  cleanup:
    return result;
}

static cpl_error_code
xsh_nod_rectify_and_shift_with_key(const char* nod_name,
                                   const char* qual_name,
                                   const char* rec_prefix,
                                   cpl_frame* frame2D,
                                   cpl_frame* order_tab_edges,
                                   cpl_frame* wave_tab,
                                   cpl_frame* model_config_frame,
                                   cpl_frame* spectral_format,
                                   cpl_frame* disp_tab_frame,
                                   xsh_instrument* instrument,
                                   xsh_rectify_param *rectify_par,
				   double** ref_ra,
				   double** ref_dec,
                                   cpl_frame **shift2_frame,
                                   cpl_frame **shift1_frame)

{

  //HERE BUG
   char file_name[80];
   cpl_frame * rec2_frame = NULL ; /**< Output of xsh_rectfiy */
   cpl_frame * rec2eso_frame = NULL ;
   cpl_frame * rec2tab_frame = NULL ;
   cpl_frame *shift2eso_frame = NULL;
   cpl_frame * rec1_frame = NULL ;
   cpl_frame *shift1eso_frame = NULL;

   /* Localization should be done following the order trace, not rectifying */
   sprintf(file_name,"REC2_%s_%s",qual_name,nod_name);
   check(rec2_frame = xsh_rectify( frame2D,order_tab_edges,
                                   wave_tab,model_config_frame,instrument, 
                                   rectify_par,spectral_format,disp_tab_frame,
                                   file_name,&rec2eso_frame,&rec2tab_frame,
                                   rec_prefix));

   sprintf(file_name,"SHIFT2_%s_%s",qual_name,nod_name);

   check( *shift2_frame = shift_with_kw( rec2_frame, instrument,rectify_par, 
                                        file_name, &shift2eso_frame, 
                                        ref_ra,ref_dec, 0));

   sprintf(file_name,"REC1_FAST_%s_%s",qual_name,nod_name);
   check( rec1_frame = xsh_rec_list_frame_invert( rec2_frame, file_name,
                                                  instrument));

   sprintf(file_name,"SHIFT1_FAST_%s_%s",qual_name,nod_name);
   check( *shift1_frame = shift_with_kw( rec1_frame, instrument,
                                        rectify_par, file_name, 
                                        &shift1eso_frame, ref_ra,
                                        ref_dec, 1));
  cleanup:

    xsh_free_frame( &rec1_frame);
    xsh_free_frame( &rec2_frame);
    xsh_free_frame( &shift1eso_frame);
    xsh_free_frame( &shift2eso_frame);
    xsh_free_frame( &rec2eso_frame);
    xsh_free_frame( &rec2tab_frame);

   return cpl_error_get_code();
}

cpl_error_code
xsh_scired_slit_nod_fast(
			 cpl_frameset *nod_set,
			 cpl_frame* spectral_format,
			 cpl_frame* master_flat,
			 cpl_frame* order_tab_edges,
                         cpl_frame* wave_tab,
                         cpl_frame* model_config_frame,
                         cpl_frame* disp_tab_frame,
                         cpl_frame* wavemap,
                         cpl_frame* slitmap,
			 xsh_instrument* instrument,
                         xsh_remove_crh_single_param *crh_single_par,
                         xsh_rectify_param *rectify_par,
                         const int do_flatfield,
                         const int compute_eff, 
                         const char* rec_prefix,
                         cpl_frameset **comb_set,
                         cpl_frameset **comb_eff_set)

{
  /* Intermediate frames */
  cpl_frame *rm_crh = NULL;
  cpl_frame * div2_frame = NULL ; 
  cpl_frame * eff2_frame = NULL ; 
  cpl_frame *shift1_frame = NULL;
  cpl_frame *shift2_frame = NULL;
  cpl_frame *shift1_eff_frame = NULL;
  cpl_frame *shift2_eff_frame = NULL;
  double* ref_ra=NULL;
  double* ref_dec=NULL;
  double* ref_eff_ra=NULL;
  double* ref_eff_dec=NULL;

  /* Output Frames (results)*/
  int i=0;
  int nb_pairs=0;

  xsh_msg("Method fast");
  check( *comb_set = cpl_frameset_new());

  if(compute_eff) {
     check( *comb_eff_set = cpl_frameset_new());
  }

  /* Now subtract B to A */
  check( nb_pairs = cpl_frameset_get_size( nod_set));

  for( i = 0; i< nb_pairs; i++){
    cpl_frame *nod_frame = NULL;
    const char* nod_name = NULL;
    char tag[80];

    nod_frame = cpl_frameset_get_frame( nod_set, i);
    nod_name = cpl_frame_get_filename( nod_frame); 
    xsh_msg_dbg_high( "***** Frame %s", nod_name);

    /* 
       No background subtraction in nodding  because bkg goes away with A-B 
    */

    /* TODO Do CRH single only if frame was not already CRH subtracted */
    /* As CRH clip works only on positive signal and nodding generate negatives
       we compute the absolute frame */
    sprintf(tag,"NOCRH_%s", xsh_instrument_arm_tostring( instrument ));

    check( rm_crh = xsh_abs_remove_crh_single( nod_frame, wavemap,instrument,
					   crh_single_par, tag));
  
    /* we now correct the absolute frame back to original layout */
    /* divide by flat */
    sprintf(tag,"DIV2_FLATFIELD_%s", xsh_instrument_arm_tostring(instrument)) ;
    check( div2_frame =xsh_frame_divide_flat( do_flatfield, rm_crh, master_flat,
					tag, instrument));

    /* Rectify */
    check(xsh_nod_rectify_and_shift_with_key(nod_name,"RECTIFIED",rec_prefix,
                                             div2_frame,order_tab_edges,
                                             wave_tab,model_config_frame,
                                             spectral_format,
                                             disp_tab_frame,
                                             instrument,
                                             rectify_par,&ref_ra,&ref_dec,
                                             &shift1_frame,&shift2_frame));

    cpl_frameset_insert( *comb_set, shift2_frame);
    cpl_frameset_insert( *comb_set, shift1_frame);

    if(compute_eff) {  
       if(disp_tab_frame != NULL) {
       int conserve_flux=rectify_par->conserve_flux;

       sprintf(tag,"NOCRH_EFF_%s",xsh_instrument_arm_tostring(instrument)) ;
       check( eff2_frame =xsh_frame_divide_flat(0, rm_crh, master_flat,
                                                tag, instrument));
       rectify_par->conserve_flux=1;
       check(xsh_nod_rectify_and_shift_with_key(nod_name,"EFF",rec_prefix,
                                                eff2_frame,order_tab_edges,
                                                wave_tab,model_config_frame,
                                                spectral_format,
                                                disp_tab_frame,
                                                instrument,
                                                rectify_par,
						&ref_eff_ra,&ref_eff_dec,
                                                &shift1_eff_frame,
                                                &shift2_eff_frame));
       rectify_par->conserve_flux=conserve_flux;  
       cpl_frameset_insert( *comb_eff_set, shift2_eff_frame);
       cpl_frameset_insert( *comb_eff_set, shift1_eff_frame);
       xsh_free_frame( &eff2_frame);

       }
    }
    xsh_free_frame( &rm_crh);
    xsh_free_frame( &div2_frame);


  } /* end loop on input pairs */

  cleanup:
    XSH_FREE( ref_ra);
    XSH_FREE( ref_dec);
    XSH_FREE( ref_eff_ra);
    XSH_FREE( ref_eff_dec);
    xsh_free_frame( &eff2_frame);
    xsh_free_frame( &div2_frame);
    return cpl_error_get_code();
}




cpl_frameset*
xsh_scired_slit_nod_accurate(
			     cpl_frameset *nod_set,
			     cpl_frame* spectral_format,
                             cpl_frame* master_flat,
                             cpl_frame* order_tab_edges,
                             cpl_frame* wave_tab,
                             cpl_frame* model_config_frame,
                             cpl_frame* disp_tab_frame,
                             cpl_frame* wavemap,
                             cpl_frame* slitmap,
                             cpl_frame *skymask_frame,
                             xsh_instrument* instrument,
                             xsh_remove_crh_single_param *crh_single_par,
                             xsh_rectify_param *rectify_par,
                             xsh_localize_obj_param *loc_obj_par,
                             xsh_slit_limit_param *slit_limit_par,
                             const char *throw_name,
                             const int do_flatfield,
                             const char* rec_prefix
			     )
{
  int i=0;
  cpl_frame * a_b = NULL;

  /* all what follows regards the old-slow method */
  char file_tag[40];
  char file_name[80];
  //int npairs=0;
  int nb_nod=0;
  double *throw_shift = NULL;
  double *throw_tab = NULL;

  cpl_frame * loc_a_b = NULL, * loc_b_a = NULL ;
  cpl_frame * loc_a_b_0 = NULL;
  
  cpl_frameset *comb_set = NULL;


  xsh_msg("Method accurate");
  check( comb_set = cpl_frameset_new());

  check( nb_nod = cpl_frameset_get_size( nod_set));
 

  if ( strcmp( throw_name, "") != 0){
    FILE* throw_file = NULL;
    char throw_line[200];
    int iline;

    xsh_msg("Load Throw shifts from file %s", throw_name);
    throw_file = fopen( throw_name, "r");

    XSH_CALLOC( throw_tab, double, nb_nod);

    iline=0;

    while (fgets( throw_line, 200, throw_file) != NULL){
      char shiftval[200];

      sscanf( throw_line, "%s",shiftval);
      if (iline < nb_nod){
        throw_tab[iline] = atof( shiftval);
      }
      iline++;
    }
    if (iline != nb_nod){
      xsh_error_msg("Invalid number of lines in ASCII file %s : %d - expected number : %d", 
        throw_name, iline, nb_nod);
    }
    fclose( throw_file);
  }

  /* For each  of frames */
  for( i = 0; i<nb_nod ; i++) {
    cpl_frame * rm_crh = NULL ;
    cpl_frame * divided = NULL ;
    cpl_frame * shifted_a_b = NULL;
    cpl_frame * shifted_a_b_eso = NULL;
    cpl_frame * shifted_a_b_tab = NULL;
    cpl_frame *inv_shifted_a_b = NULL;
    cpl_frame *loc_rec = NULL;
    cpl_frame *loc_rec_eso = NULL;
    cpl_frame *loc_rec_tab = NULL;
    cpl_frame *inv_loc_rec = NULL;
    char str[16] ;
    int mode =0;

    if (throw_tab != NULL){
      throw_shift = &throw_tab[i];
    }

    if (i%2 == 0){
      sprintf( str, "AB_%d_%s", i/2, xsh_instrument_arm_tostring(instrument)) ;
      mode =0;
    }
    else{
      sprintf( str, "BA_%d_%s", i/2, xsh_instrument_arm_tostring(instrument)) ;
      mode=1;
    }
    check( a_b = cpl_frameset_get_frame( nod_set, i));
    xsh_msg_dbg_high( "***** Frame %s", cpl_frame_get_filename( a_b ) ) ;

    /* remove_crh : if we have done multiple don't do single */
    xsh_msg( "Remove crh (single frame)" ) ;
    sprintf(file_tag, "SLIT_NOD_NOCRH_%s",str) ;

    check( rm_crh = xsh_abs_remove_crh_single(a_b, wavemap, instrument,
      crh_single_par, file_tag));

    /*divide by flat field */
    sprintf(file_tag,"SLIT_NOD_NOCRH_FF_%s", str);
    check( divided = xsh_frame_divide_flat( do_flatfield, rm_crh,master_flat, file_tag,
      instrument));

    if ( i > 1 && throw_shift == NULL){
      if (i%2 == 0){
        xsh_msg("TEST %d : on fait loc_ab et loc_ba", i);
        sprintf(file_name,"LOC_REC_%s.fits", str);
        check( loc_rec = xsh_rectify( divided, order_tab_edges,
                                    wave_tab, model_config_frame, instrument,
                                    rectify_par, spectral_format, disp_tab_frame,
                                    file_name, &loc_rec_eso, &loc_rec_tab, rec_prefix));
        xsh_msg( "Localize the object (A%d)", i/2);
        sprintf( file_name,"LOCALIZE_%s.fits", str);
        check( loc_a_b = xsh_localize_obj( loc_rec, skymask_frame, instrument, loc_obj_par,
          NULL, file_name));

        xsh_msg( "Negative rectify");
        sprintf(file_name,"DRL_INV_LOC_REC_%s", str);
      check( inv_loc_rec = xsh_rec_list_frame_invert( loc_rec,
        file_name, instrument));

      xsh_msg( "Localize the object (B%d)", i/2);
      sprintf(file_name,"INV_LOCALIZE_%s.fits", str);
      check( loc_b_a = xsh_localize_obj( inv_loc_rec, skymask_frame, instrument, loc_obj_par,
        NULL, file_name));
      }
      else{
        xsh_msg("TEST %d : loc_ab = loc_ba", i);
        loc_a_b = loc_b_a;
      }
    }
    /* Rectify and shift */
    xsh_msg( "Rectify and shift");
    sprintf(file_name,"SHIFT_REC_NOCRH_%s.fits", str);

    check( shifted_a_b = xsh_rectify_and_shift( divided,order_tab_edges,
                                                 wave_tab,
                                                 model_config_frame,
                                                 instrument,
                                                 rectify_par,
                                                 spectral_format,
                                                 loc_a_b,
                                                 loc_a_b_0,
                                                 throw_shift,
                                                 disp_tab_frame,
                                                 file_name,
                                                 &shifted_a_b_eso,
                                                 &shifted_a_b_tab));
    xsh_free_frame( &loc_a_b);

    if ( i==0 && throw_shift == NULL){
      xsh_msg( "Localize the object (A0)");
      sprintf(file_name,"LOCALIZE_%s.fits", str);
      check( loc_a_b_0 = xsh_localize_obj( shifted_a_b, skymask_frame, instrument, loc_obj_par,
        NULL, file_name));
  
      xsh_msg( "Negative rectify");
      sprintf(file_name,"DRL_INV_LOC_REC_%s", str);
      check( inv_shifted_a_b = xsh_rec_list_frame_invert( shifted_a_b, 
        file_name, instrument));
      xsh_msg( "Localize the object (B0)");
      sprintf(file_name,"INV_LOCALIZE_%s.fits", str);
      check( loc_a_b = xsh_localize_obj( inv_shifted_a_b, skymask_frame, instrument, loc_obj_par,
        NULL, file_name));
    }
    check( cpl_frameset_insert( comb_set, shifted_a_b));

    xsh_free_frame( &rm_crh);
    xsh_free_frame( &divided);
    xsh_free_frame( &shifted_a_b_eso);
    xsh_free_frame( &shifted_a_b_tab);
    xsh_free_frame( &inv_shifted_a_b);
    xsh_free_frame( &loc_rec);
    xsh_free_frame( &loc_rec_eso);
    xsh_free_frame( &loc_rec_tab);
    xsh_free_frame( &inv_loc_rec);
  } /* end loop on each pair */

  cleanup:
    if ( cpl_error_get_code() != CPL_ERROR_NONE){
      xsh_free_frameset( &comb_set);
    }
    xsh_free_frame( &loc_a_b_0);
    xsh_free_frame( &loc_a_b);
    XSH_FREE( throw_tab);
    return comb_set;


}


void
xsh_scired_slit_nod_get_calibs(cpl_frameset* raws,
                               cpl_frameset* calib,
			       xsh_instrument* instrument,
			       cpl_frame** bpmap,
			       cpl_frame** master_bias,
			       cpl_frame** master_flat,
			       cpl_frame** order_tab_edges,
			       cpl_frame** wave_tab,
			       cpl_frame** model_config_frame,
			       cpl_frame** wavemap,
			       cpl_frame** slitmap,
			       cpl_frame** disp_tab_frame,
			       cpl_frame** spectral_format,
                               cpl_frame** skymask_frame,
                               cpl_frame** response_ord_frame,
                               cpl_frame** frm_atmext,
                               int do_computemap,
                               int use_skymask,
			       const char* rec_prefix
			       )

{
  int recipe_use_model=FALSE;
  *bpmap = xsh_find_master_bpmap( calib);
  cpl_frame* ref_frame=NULL;

  /* In UVB and VIS mode */
  if ( xsh_instrument_get_arm(instrument) != XSH_ARM_NIR){
    /* RAWS frameset must have only one file */
    check(*master_bias = xsh_find_master_bias( calib, instrument));
  }
 
  check( *order_tab_edges = xsh_find_order_tab_edges( calib, instrument));

  /* one should have either model config frame or wave sol frame */
  if((*model_config_frame = xsh_find_frame_with_tag(calib,
                                                    XSH_MOD_CFG_OPT_2D,
                                                    instrument)) == NULL) {

     xsh_error_reset();
     if ((*model_config_frame = xsh_find_frame_with_tag(calib,XSH_MOD_CFG_TAB, 
                                               instrument)) == NULL) {
        xsh_error_reset();

        if ((*model_config_frame = xsh_find_frame_with_tag(calib,XSH_MOD_CFG_OPT_AFC, 
                                                  instrument)) == NULL) {
           xsh_error_reset();
        }
     }

  }
  if ( *model_config_frame == NULL){
    xsh_msg("RECIPE USE WAVE SOLUTION");
    check( *wave_tab = xsh_find_wave_tab( calib, instrument));
    recipe_use_model = FALSE;
  }
  else{
    xsh_msg("RECIPE USE MODEL");
    recipe_use_model = TRUE;
  }
  cpl_error_reset() ;

  XSH_ASSURE_NOT_ILLEGAL( model_config_frame != NULL || wave_tab != NULL);

  check( *master_flat = xsh_find_master_flat( calib, instrument));
  if(do_computemap && recipe_use_model==FALSE) {
     check( *wavemap = xsh_find_wavemap( calib, instrument));
  } 
  /* to compute efficiency */
  if(NULL == (*disp_tab_frame = xsh_find_disp_tab( calib, instrument))) {
     xsh_msg("To compute efficiency, you must provide a DISP_TAB_ARM input");
  }

  if ( recipe_use_model){
    ref_frame=cpl_frameset_get_first(raws);
  }
  else{
    ref_frame = *master_flat;
  }
  check( xsh_check_get_map( *disp_tab_frame, *order_tab_edges,
    ref_frame, *model_config_frame, calib, instrument,
    do_computemap, recipe_use_model, rec_prefix,
    wavemap, slitmap));

  check( *spectral_format = xsh_find_spectral_format( calib, instrument));

  if ( use_skymask == TRUE){
    xsh_msg("Using sky mask");
    check( *skymask_frame = xsh_find_frame_with_tag( calib, XSH_SKY_LINE_LIST,
      instrument));
  }

  check( *response_ord_frame = xsh_find_frame_with_tag(calib,
                                                   XSH_RESPONSE_MERGE1D_SLIT,
                                                  instrument));

  if(*response_ord_frame != NULL) { 
     *frm_atmext=xsh_find_frame_with_tag(calib,XSH_ATMOS_EXT,instrument);
     if(*frm_atmext==NULL) {
        xsh_msg_error("Provide atmospheric extinction frame");
     }
  }

  cleanup:
    return;
}

/** 
  @brief set parameter defaults
  @param pars input parameters
  @param inst instrument arm and lamp setting
  @param rectify_par rectify parameter structure
 */
static cpl_error_code 
xsh_scired_nod_params_set_defaults(cpl_parameterlist* pars,
                                   xsh_instrument* inst,
                                   xsh_rectify_param * rectify_par,
                                   const char* rec_id)
{
   cpl_parameter* p=NULL;
 
  if (xsh_instrument_get_arm(inst) == XSH_ARM_NIR){

     check(p=xsh_parameters_find(pars,rec_id,"rectify-bin-slit"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        rectify_par->rectif_bin_space=XSH_SLIT_BIN_SIZE_PIPE_NIR;
        cpl_parameter_set_double(p,XSH_SLIT_BIN_SIZE_PIPE_NIR);
     }

     check(p=xsh_parameters_find(pars,rec_id,"rectify-bin-lambda"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        rectify_par->rectif_bin_lambda=XSH_WAVE_BIN_SIZE_PIPE_NIR;
        cpl_parameter_set_double(p,XSH_WAVE_BIN_SIZE_PIPE_NIR);
     }



  } else {


     check(p=xsh_parameters_find(pars,rec_id,"rectify-bin-slit"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        rectify_par->rectif_bin_space=XSH_SLIT_BIN_SIZE_PIPE_UVB;
        cpl_parameter_set_double(p,XSH_SLIT_BIN_SIZE_PIPE_UVB);
     }

     check(p=xsh_parameters_find(pars,rec_id,"rectify-bin-lambda"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        rectify_par->rectif_bin_lambda=XSH_WAVE_BIN_SIZE_PIPE_UVB;
        cpl_parameter_set_double(p,XSH_WAVE_BIN_SIZE_PIPE_UVB);
     }


  }

 cleanup:
 
  return cpl_error_get_code();

}


/* for the moment not used */
/** 
  @brief monitor param values
  @param rectify_par rectify param structure
  @param loc_obj_par localizzation object param
 */
static cpl_error_code 
xsh_scired_nod_params_monitor(xsh_rectify_param * rectify_par,
                   xsh_localize_obj_param * loc_obj_par)
{



  xsh_msg_dbg_low("rectify params: radius=%g bin_lambda=%g bin_space=%g",
	  rectify_par->rectif_radius,rectify_par->rectif_bin_lambda,
	  rectify_par->rectif_bin_space);

  xsh_msg_dbg_low("localize params: chunk_nb=%d nod_step=%g",
	  loc_obj_par->loc_chunk_nb,loc_obj_par->nod_step);

  return cpl_error_get_code();

}




cpl_error_code
xsh_scired_nod_get_parameters(cpl_parameterlist* parameters,
			      xsh_instrument* instrument,
			      xsh_remove_crh_single_param** crh_single_par,
			      xsh_rectify_param** rectify_par,
			      xsh_merge_param** merge_par,
			      xsh_extract_param** extract_par,
			      xsh_combine_nod_param** combine_nod_par,
			      xsh_slit_limit_param** slit_limit_par,
			      xsh_localize_obj_param** loc_obj_par,
			      int* rectify_fast,
                              int binx,int biny,const char* rec_id
			      )


{
  check( *loc_obj_par = xsh_parameters_localize_obj_get(rec_id,
							parameters));
  check( *rectify_par = xsh_parameters_rectify_get(rec_id,
						   parameters));
  check( *rectify_fast = xsh_parameters_rectify_fast_get(rec_id,
							 parameters));
  check( *crh_single_par = xsh_parameters_remove_crh_single_get(rec_id,
								parameters));
  check( *merge_par = xsh_parameters_merge_ord_get(rec_id,
						   parameters));
  check( *extract_par=xsh_parameters_extract_get(rec_id, parameters));
  check( *combine_nod_par = xsh_parameters_combine_nod_get(rec_id,
							  parameters)) ;
  check( *slit_limit_par = xsh_parameters_slit_limit_get(rec_id,
							 parameters));
  check(xsh_scired_nod_params_set_defaults(parameters,instrument,*rectify_par,rec_id));

  /* adjust relevant parameter to binning: for the moment not used 
  if ( xsh_instrument_get_arm( instrument ) != XSH_ARM_NIR ) {
    check(xsh_params_bin_scale(binx,biny));
  }
  */

  check( xsh_scired_nod_params_monitor(*rectify_par,*loc_obj_par));

  cleanup:
    return cpl_error_get_code();
}



cpl_error_code
xsh_flux_calibrate(cpl_frame* rect2D,
                   cpl_frame* rect1D,
                   cpl_frame* atmext,
                   cpl_frame* response,
                   xsh_merge_param * mpar,
                   xsh_instrument* inst,
                   const char* rec_prefix,
                   cpl_frame** fcal_rect_2D,
                   cpl_frame** fcal_rect_1D,
                   cpl_frame** fcal_2D,
                   cpl_frame** fcal_1D)
{
  char file_tag[80];
  char arm_str[8] ;
  cpl_frame* nrm_2D=NULL;
  cpl_frame* nrm_1D=NULL;

  sprintf(arm_str,"%s",xsh_instrument_arm_tostring(inst));
  
  sprintf(file_tag,"%s_%s_%s",rec_prefix,XSH_NORM_ORDER2D,arm_str);
  check(nrm_2D=xsh_normalize_spectrum_ord(rect2D,atmext,0,inst,file_tag));
  sprintf(file_tag,"%s_%s_%s",rec_prefix,XSH_FLUX_ORDER2D,arm_str);
  check(*fcal_rect_2D=xsh_util_multiply_by_response_ord(nrm_2D,response,file_tag));
  check(*fcal_2D = xsh_merge_ord(*fcal_rect_2D, inst,mpar,rec_prefix ));

  
  sprintf(file_tag,"%s_%s_%s",rec_prefix,XSH_NORM_ORDER1D,arm_str);
  check(nrm_1D=xsh_normalize_spectrum_ord(rect1D,atmext,0,inst,file_tag));
  sprintf(file_tag,"%s_%s_%s",rec_prefix,XSH_FLUX_ORDER1D,arm_str);
  check(*fcal_rect_1D=xsh_util_multiply_by_response_ord(nrm_1D,response,file_tag));
  check(*fcal_1D= xsh_merge_ord(*fcal_rect_1D,inst,mpar,rec_prefix ));
 	
 cleanup:
  xsh_free_frame(&nrm_2D);
  xsh_free_frame(&nrm_1D);
 
  return cpl_error_get_code();
}


cpl_error_code
xsh_slit_stare_get_calibs(cpl_frameset* calib, 
                          xsh_instrument* instrument,
                          cpl_frame** spectralformat,
                          cpl_frame** mbias,
                          cpl_frame** mdark,
                          cpl_frame** mflat,
                          cpl_frame** otab_edges,
                          cpl_frame** model_cfg,
                          cpl_frame** wave_tab,
                          cpl_frame** sky_list,
                          cpl_frame** qc_sky,
                          cpl_frame** bpmap,
                          cpl_frame** sframe_sky_sub_tab,
                          cpl_frame** wmap,
                          cpl_frame** smap,
                          int * recipe_use_model)
{


   int binx=1;
   int biny=1;
   cpl_propertylist* plist=NULL;

   check(*spectralformat =xsh_find_frame_with_tag(calib,XSH_SPECTRAL_FORMAT,
                                                  instrument));

   XSH_ASSURE_NOT_NULL_MSG(*spectralformat,"Null input spectral format frame");
   check(xsh_instrument_update_from_spectralformat(instrument,*spectralformat));

   /* In UVB and VIS mode */
   if ( xsh_instrument_get_arm(instrument) != XSH_ARM_NIR){
      /* RAWS frameset must have only one file */
      if((*mbias = xsh_find_master_bias( calib, instrument)) == NULL) {
         xsh_msg_error("You must give a MASTER_BIAS_ARM frame");
         return CPL_ERROR_DATA_NOT_FOUND;
      }
   }

   if((*mdark =  xsh_find_master_dark( calib, instrument))==NULL) {
      xsh_msg_warning("Frame %s not provided",XSH_MASTER_DARK);
      xsh_error_reset();
   }

   check( *mflat = xsh_find_master_flat( calib, instrument ));
   if ( xsh_instrument_get_arm(instrument) != XSH_ARM_NIR) {
      check(plist=cpl_propertylist_load(cpl_frame_get_filename(*mflat),0));
      check(binx=xsh_pfits_get_binx(plist));
      check(biny=xsh_pfits_get_biny(plist));
      xsh_free_propertylist(&plist);
   }
   check(*otab_edges = xsh_find_order_tab_edges(calib,instrument));

   /* one should have either model config frame or wave sol frame */
   if((*model_cfg = xsh_find_frame_with_tag(calib,XSH_MOD_CFG_OPT_2D,
                                            instrument)) == NULL) {
      xsh_error_reset();
      if ((*model_cfg = xsh_find_frame_with_tag(calib,XSH_MOD_CFG_TAB, 
                                                instrument)) == NULL) {
         xsh_error_reset();

         if ((*model_cfg = xsh_find_frame_with_tag(calib,XSH_MOD_CFG_OPT_AFC, 
                                                   instrument)) == NULL) {
            xsh_error_reset();
         }
      }
   }
   xsh_error_reset();
   cpl_error_reset();

   if ( *model_cfg == NULL){
      xsh_msg("RECIPE USE WAVE SOLUTION");
      check( *wave_tab = xsh_find_wave_tab( calib, instrument));
      *recipe_use_model = FALSE;
   }
   else{
      xsh_msg("RECIPE USE MODEL");
      *recipe_use_model = TRUE;
   }

   if ( *model_cfg != NULL){
      check(*sky_list = xsh_find_frame_with_tag( calib, XSH_SKY_LINE_LIST,
                                                 instrument));
      if ( *sky_list != NULL){
         /* we compute unbinned THE map to easier plots from QC garching */
         check(*qc_sky=xsh_util_physmod_model_THE_create(*model_cfg,instrument,
                                                         *sky_list,1,1,9,1));
      }
   }

   /* optional frames */
   *bpmap = xsh_find_master_bpmap(calib);
   if ((*sframe_sky_sub_tab = xsh_find_frame_with_tag(calib,XSH_SKY_SUB_BKPTS,
                                                      instrument)) == NULL) {
      xsh_error_reset();
   } else {
      xsh_msg_warning("Data reduction with user defined break points number from file %s",
                      cpl_frame_get_filename(*sframe_sky_sub_tab));
   }


   check(*wmap = xsh_find_frame_with_tag(calib,XSH_WAVE_MAP,instrument));
   check(*smap = xsh_find_frame_with_tag(calib,XSH_SLIT_MAP,instrument));

  cleanup:
   xsh_free_propertylist(&plist);
   return cpl_error_get_code();
}

cpl_error_code
xsh_slit_offset_get_calibs(cpl_frameset* calib,xsh_instrument* instrument,
                           cpl_frame** bpmap,cpl_frame** mbias,
                           cpl_frame** mdark, cpl_frame** otab_edges,
                           cpl_frame** model_cfg, cpl_frame** wave_tab,
                           cpl_frame** mflat, cpl_frame** wmap, 
                           cpl_frame** spectral_format)
{



   check(*bpmap = xsh_find_master_bpmap( calib));

   /* In UVB and VIS mode */
   if ( xsh_instrument_get_arm(instrument) != XSH_ARM_NIR){
      /* RAWS frameset must have only one file */
      check( *mbias = xsh_find_frame_with_tag(calib,XSH_MASTER_BIAS,instrument));
   }

   /* TODO: Is mdark input really needed? */
   if((*mdark=xsh_find_frame_with_tag(calib,XSH_MASTER_DARK,instrument))==NULL){
      xsh_error_reset();
   }


   check( *otab_edges = xsh_find_order_tab_edges( calib, instrument));
   if((*model_cfg=xsh_find_frame_with_tag(calib,XSH_MOD_CFG_OPT_2D,instrument)) == NULL) {
      xsh_error_reset();
      if ((*model_cfg = xsh_find_frame_with_tag(calib,XSH_MOD_CFG_TAB, 
                                                instrument)) == NULL) {
         xsh_error_reset();

         if ((*model_cfg = xsh_find_frame_with_tag(calib,XSH_MOD_CFG_OPT_AFC, 
                                                   instrument)) == NULL) {
            xsh_error_reset();
         }
      }
   }
   cpl_error_reset() ;

   *wave_tab = xsh_find_wave_tab( calib, instrument ) ;
   cpl_error_reset() ;
   XSH_ASSURE_NOT_ILLEGAL( *model_cfg != NULL || *wave_tab != NULL ) ;

   check( *mflat = xsh_find_master_flat( calib, instrument));

   check( *wmap = xsh_find_wavemap( calib, instrument));
   check( *spectral_format = xsh_find_spectral_format( calib, instrument));

 
  cleanup:
   return cpl_error_get_code();

}

cpl_error_code
xsh_slit_stare_get_params(cpl_parameterlist* parameters,
                          const char* rec_id,
                          int* pre_overscan_corr,
                          xsh_clipping_param** crh_clipping_par,
                          xsh_background_param** backg_par,
                          xsh_localize_obj_param** loc_obj_par,
                          xsh_rectify_param** rectify_par,
                          xsh_remove_crh_single_param** crh_single_par,
                          int* sub_sky_nbkpts1,
                          int* do_flatfield,
                          int* sub_sky_nbkpts2,
                          xsh_subtract_sky_single_param** sky_par,
                          xsh_extract_param** extract_par,
                          int* do_optextract,
                          xsh_merge_param** merge_par,
                          xsh_opt_extract_param** opt_extract_par,
                          int* do_trace_obj)
{

   check( *pre_overscan_corr = xsh_parameters_get_int( parameters, rec_id,
                                                       "pre-overscan-corr"));

   check( *crh_clipping_par = xsh_parameters_clipping_crh_get(rec_id,parameters));
   check( *backg_par = xsh_parameters_background_get(rec_id,parameters));
 
   check( *loc_obj_par = xsh_parameters_localize_obj_get(rec_id,parameters));
   check( *rectify_par = xsh_parameters_rectify_get(rec_id,parameters));
   check( *crh_single_par = xsh_parameters_remove_crh_single_get(rec_id,parameters));

   check( *sub_sky_nbkpts1 = xsh_parameters_subtract_sky_single_get_first(
             rec_id,parameters));
   check( *do_flatfield = xsh_parameters_get_boolean( parameters, rec_id,
                                                      "do-flatfield"));
   check( *sub_sky_nbkpts2 = xsh_parameters_subtract_sky_single_get_second(
             rec_id,parameters));
   check( *sky_par= xsh_parameters_subtract_sky_single_get(rec_id,parameters));

   check( *extract_par = xsh_parameters_extract_get( rec_id, parameters));
   check( *do_optextract = xsh_parameters_get_boolean( parameters, rec_id,
                                                       "do-optextract"));
   check( *opt_extract_par = xsh_parameters_opt_extract_get(rec_id,parameters));

   check( *merge_par = xsh_parameters_merge_ord_get(rec_id,parameters));

   check( *do_trace_obj = xsh_parameters_get_boolean( parameters,rec_id,
                                                      "trace-obj"));


  cleanup:
   return cpl_error_get_code();
}


cpl_error_code
xsh_slit_offset_get_params(cpl_parameterlist* parameters,
                          const char* rec_id,
                          xsh_clipping_param** crh_clipping_par,
                          xsh_background_param** backg_par,
                          xsh_localize_obj_param** loc_obj_par,
                          xsh_rectify_param** rectify_par,
                          xsh_remove_crh_single_param** crh_single_par,
                          double* opt_kappa,
                          xsh_merge_param** merge_par,
                          xsh_extract_param**extract_par,
                          xsh_combine_nod_param** combine_nod_param,
                          int* do_flatfield,
                          int* gen_sky)
{


   check( *crh_clipping_par = xsh_parameters_clipping_crh_get(rec_id,parameters));
   check( *backg_par = xsh_parameters_background_get(rec_id,parameters));

   check( *loc_obj_par = xsh_parameters_localize_obj_get(rec_id,parameters));
   check( *rectify_par = xsh_parameters_rectify_get(rec_id,parameters));
   check( *crh_single_par = xsh_parameters_remove_crh_single_get(rec_id,parameters));
   if ( (*rectify_par)->rectify_full_slit == 1 ){

      xsh_msg( "Use Full Slit" ) ;
   }
   else {
      xsh_msg( "Use Max Possible Slit" ) ;
   }

   check( *opt_kappa = xsh_parameters_optimal_extract_get_kappa(rec_id,parameters));

   check( *merge_par = xsh_parameters_merge_ord_get(rec_id,parameters));

   check(*extract_par=xsh_parameters_extract_get(rec_id, parameters )) ;

   check(*combine_nod_param = xsh_parameters_combine_nod_get(rec_id,parameters )) ;

   check( *do_flatfield = xsh_parameters_get_boolean( parameters,rec_id,
                                                      "do-flatfield"));

   check( *gen_sky = xsh_parameters_get_boolean( parameters, rec_id,
                                                 "gen-sky"));

  cleanup:
   return cpl_error_get_code();

}


cpl_error_code
xsh_slit_stare_correct_crh_and_sky(xsh_localize_obj_param * loc_obj_par,
                                   xsh_remove_crh_single_param* crh_single_par,
                                   xsh_rectify_param* rectify_par,
                                   int do_sub_sky,
                                   const char* rec_prefix,
                                   cpl_frame* rmbkg,
                                   cpl_frame* order_tab_edges, 
                                   cpl_frame* slitmap, 
                                   cpl_frame* wavemap,
                                   cpl_frame* model_config,
                                   cpl_frame* single_frame_sky_sub_tab,
                                   xsh_instrument* instrument,
                                   int sub_sky_nbkpts1, 
                                   xsh_subtract_sky_single_param* sky_par,
                                   cpl_frame** sky,
                                   cpl_frame** sky_eso,
                                   cpl_frame** sky_ima,
                                   cpl_frame* wave_tab,
                                   cpl_frame* disp_tab,
                                   cpl_frame* spectral_format,
                                   int nb_raw_frames,
                                   cpl_frame** loc_table,
                                   cpl_frame** clean,
                                   cpl_frame** clean_obj)
{

  char sky_tag[80];
  char rec_name[80];
  cpl_frame * sub_sky = NULL ; /**< Output of xsh_subtract_sky_single */
  cpl_frame * rect = NULL ; /**< Output of xsh_rectfiy */

  if( (loc_obj_par->method != LOC_MANUAL_METHOD) || 
         (nb_raw_frames == 1 && crh_single_par->nb_iter > 0)){
     xsh_msg("Preliminary sky subtraction for CRH single or automatic localization");
     xsh_msg("Sky will be put back later");

    sprintf(sky_tag,"%s_TMPSKY",rec_prefix);
    if(do_sub_sky) {
      check( sub_sky = xsh_check_subtract_sky_single( CPL_TRUE, rmbkg, 
							    order_tab_edges, 
                                                            slitmap, 
                                                            wavemap, 
                                                            NULL, NULL,
							    single_frame_sky_sub_tab, 
                                                            instrument, 
                                                            sub_sky_nbkpts1, 
                                                            sky_par,
							    sky, 
                                                            sky_eso, 
                                                            sky_ima, 
                                                            sky_tag));
    } else {
      sub_sky = cpl_frame_duplicate(rmbkg);
    }
     if( loc_obj_par->method != LOC_MANUAL_METHOD){
        xsh_msg("Localize auto");
        sprintf(rec_name,"%s_%s_%s.fits",
		rec_prefix,XSH_ORDER2D,
                xsh_instrument_arm_tostring(instrument));

        /* rectify */ 
  
        check( rect = xsh_rectify( sub_sky, order_tab_edges,
                                         wave_tab, model_config, 
					 instrument, rectify_par,
                                         spectral_format, disp_tab, 
                                         rec_name, NULL, NULL,rec_prefix));

        /* localize */
        check( *loc_table = xsh_localize_obj( rect, NULL, instrument,
                                                   loc_obj_par, NULL, NULL));
      xsh_free_frame( &rect);
     }
     else{
       check( *loc_table = xsh_localize_obj( NULL, NULL,
      instrument, loc_obj_par, NULL, NULL));
     }

    /* Remove COSMICS single */
    check( *clean = xsh_check_remove_crh_single( nb_raw_frames, sub_sky,
      crh_single_par, wavemap, instrument, rec_prefix));
    if(do_sub_sky) {
      check( *clean_obj = xsh_add_sky_model( *clean, *sky_ima,
					    instrument, rec_prefix));
    } else {
      *clean_obj=cpl_frame_duplicate(*clean);
    }
    
    xsh_free_frame( &sub_sky);
    xsh_free_frame( sky);
    xsh_free_frame( sky_eso);
    xsh_free_frame( sky_ima);
  } 
  else {
     /* LOcalize manual && no CRH SINGLE */
     check( *clean_obj=cpl_frame_duplicate(rmbkg));
     check( *loc_table = xsh_localize_obj( NULL, NULL,
      instrument, loc_obj_par, NULL, NULL));
  }


 cleanup:

   xsh_free_frame(&sub_sky) ;
   xsh_free_frame(&rect) ;

   return cpl_error_get_code();

}


cpl_error_code
xsh_slit_stare_get_maps(cpl_frameset* calib,
                        int do_compute_map,int recipe_use_model,
                        const char* rec_prefix,xsh_instrument* instrument,
                        cpl_frame* model_config_frame,cpl_frame* crhm_frame,
                        cpl_frame* disp_tab_frame, cpl_frame* order_tab_edges,
                        cpl_frame** wavemap_frame, cpl_frame** slitmap_frame)
{
   char wave_map_tag[80];
   char slit_map_tag[80];
   int found_temp=1;

   if ( do_compute_map){
      if (recipe_use_model){
         sprintf(wave_map_tag,"%s_%s_%s",rec_prefix,XSH_WAVE_MAP_MODEL,
                 xsh_instrument_arm_tostring( instrument ));
         sprintf(slit_map_tag,"%s_%s_%s",rec_prefix,XSH_SLIT_MAP_MODEL,
                 xsh_instrument_arm_tostring( instrument ));
      
         check(xsh_model_temperature_update_frame(&model_config_frame,crhm_frame,
                                                  instrument,&found_temp));
    
         check( xsh_create_model_map( model_config_frame, instrument,
                                      wave_map_tag,slit_map_tag, 
                                      wavemap_frame, slitmap_frame));
      }
      else{
         xsh_msg("Compute the wave map and the slit map");
         check( xsh_create_map( disp_tab_frame, order_tab_edges,
                                crhm_frame, instrument, 
                                wavemap_frame, slitmap_frame,
                                rec_prefix));
      }
   }
   else {
      xsh_msg( "Get the wave map and the slit map from sof" ) ;
      check( *slitmap_frame = xsh_find_slitmap(calib, instrument));
      check( *wavemap_frame = xsh_find_wavemap(calib, instrument));
   }

  cleanup:
   return cpl_error_get_code();
}


cpl_error_code
xsh_scired_util_spectra_flux_calibrate(cpl_frame* res2D,cpl_frame* res1D,
                                       cpl_frame* response,cpl_frame* atmext,
                                       xsh_instrument* inst,
                                       const char* prefix,
                                       cpl_frame** fluxcal_2D,
                                       cpl_frame** fluxcal_1D)
{
   char  file_tag[40];
   cpl_frame* nrm_1D=NULL;
   cpl_frame* nrm_2D=NULL;

   sprintf(file_tag,"%s_NORM2D_%s",prefix,xsh_instrument_arm_tostring(inst));
   check(nrm_2D=xsh_normalize_spectrum(res2D,atmext,0,inst,file_tag));
   sprintf(file_tag,"%s_FLUXCAL2D_%s",prefix,xsh_instrument_arm_tostring(inst));
   check(*fluxcal_2D=xsh_util_multiply_by_response(nrm_2D,response,file_tag));
   sprintf(file_tag,"%s_NORM1D_%s",prefix,xsh_instrument_arm_tostring(inst));
   check(nrm_1D=xsh_normalize_spectrum(res1D,atmext,0,inst,file_tag));
   sprintf(file_tag,"%s_FLUXCAL1D_%s",prefix,xsh_instrument_arm_tostring(inst));
   check(*fluxcal_1D=xsh_util_multiply_by_response(nrm_1D,response,file_tag));

  cleanup:
   xsh_free_frame(&nrm_1D);
   xsh_free_frame(&nrm_2D);

   return cpl_error_get_code();

}

